"""
Projections and grids
"""
# Python 2 stuff
from __future__ import division
from six import string_types

# Builtins
import copy
import warnings
from functools import partial
from distutils.version import LooseVersion

# External libs
import pyproj
import numpy as np
from scipy.interpolate import RegularGridInterpolator, RectBivariateSpline

try:
    from osgeo import osr
    has_gdal = True
except ImportError:
    has_gdal = False

# Locals
from salem import lazy_property, wgs84

try:
    crs_type = pyproj.crs.CRS
except AttributeError:
    class Dummy():
        pass
    crs_type = Dummy

def check_crs(crs, raise_on_error=False):
    """Checks if the crs represents a valid grid, projection or ESPG string.

    Examples
    --------
    >>> p = check_crs('epsg:26915 +units=m')
    >>> p.srs
    'epsg:26915 +units=m'
    >>> p = check_crs('wrong')
    >>> p is None
    True

    Returns
    -------
    A valid crs if possible, otherwise None
    """

    try:
        crs = crs.salem.grid  # try xarray
    except:
        pass

    if isinstance(crs, string_types):
        # necessary for python 2
        crs = str(crs)

    err1, err2 = None, None

    if isinstance(crs, pyproj.Proj) or isinstance(crs, Grid):
        out = crs
    elif isinstance(crs, crs_type):
        out = pyproj.Proj(crs.to_wkt(), preserve_units=True)
    elif isinstance(crs, dict) or isinstance(crs, string_types):
        if isinstance(crs, string_types):
            # quick fix for https://github.com/pyproj4/pyproj/issues/345
            crs = crs.replace(' ', '').replace('+', ' +')

        # A series of try-catch to handle the (too) many changes in pyproj
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            warnings.filterwarnings('ignore', category=FutureWarning)
            try:
                out = pyproj.Proj(crs, preserve_units=True)
            except RuntimeError as e:
                err1 = str(e)
                try:
                    out = pyproj.Proj(init=crs, preserve_units=True)
                except RuntimeError as e:
                    err2 = str(e)
                    out = None
    else:
        out = None

    if raise_on_error and out is None:
        msg = ('salem could not properly parse the provided coordinate '
               'reference system (crs). This could be due to errors in your '
               'data, in PyProj, or with salem itself. If this occurs '
               'unexpectedly, report an issue to https://github.com/fmaussion/'
               'salem/issues. Full log: \n'
               'crs: {} ; \n'.format(crs))
        if err1 is not None:
            msg += 'Output of `pyproj.Proj(crs, preserve_units=True)`: {} ; \n'
            msg = msg.format(err1)
        if err2 is not None:
            msg += 'Output of `pyproj.Proj(init=crs, preserve_units=True)`: {}'
            msg = msg.format(err2)
        raise ValueError(msg)

    return out


class Grid(object):
    """A structured grid on a map projection.

    Central class in the library, taking over user concerns about the
    gridded representation of georeferenced data. It adds a level of
    abstraction by defining a new coordinate system.

    A grid requires georeferencing information at instantiation and is
    immutable *in principle*. I didn't implement barriers: if you want to mess
    with it, you can (not recommended). Note that in most cases, users won't
    have to define the grid themselves: most georeferenced datasets contain
    enough metadata for Salem to determine the data's grid automatically.

    A grid is defined by a projection, a reference point in this projection,
    a grid spacing and a number of grid points. The grid can be defined in a
    "upper left" convention (reference point at the top left corner,
    dy negative - always -). This is the python convention, but not that of all
    datasets (for example, the output of the WRF model follows a down-left
    corner convention). Therefore, grids can also be defined in a "lower left"
    corner convention (dy positive). The use of one or the other convention
    depends on the data, so the user should take care of what he is doing.

    The reference points of the grid points might be located at the corner of
    the pixel (upper left corner for example if dy is negative) or the center
    of the pixel (most atmospheric datasets follow this convention). The two
    concepts are truly equivalent and each grid instance gives access to one
    representation or another ("center_grid" and "corner_grid" properties).
    Under the hood, Salem uses the representation it needs to do
    the job by accessing either one or the other of these parameters. The user
    should know which convention he needs for his purposes: some grid
    functions and properties are representation dependant (transform,
    ll_coordinates, ...) while some are not (extent,
    corner_ll_coordinates ...).

    Attributes
    ----------

    proj
    nx
    ny
    dx
    dy
    x0
    y0
    origin
    pixel_ref
    x_coord
    y_coord
    xy_coordinates
    ll_coordinates
    xstagg_xy_coordinates
    ystagg_xy_coordinates
    xstagg_ll_coordinates
    ystagg_ll_coordinates
    center_grid
    corner_grid
    extent
    """

    def __init__(self, proj=wgs84, nxny=None, dxdy=None, x0y0=None,
                 pixel_ref='center',
                 corner=None, ul_corner=None, ll_corner=None):
        """
        Parameters
        ----------
        proj : pyproj.Proj instance
            defines the grid's map projection. Defaults to 'PlateCarree'
            (wgs84)
        nxny : (int, int)
            (nx, ny) number of grid points
        dxdy : (float, float)
            (dx, dy) grid spacing in proj coordinates. dx must be positive,
            while dy can be positive or negative depending on the origin
            grid point's lecation (upper-left or lower-left)
        x0y0 : (float, float)
            (x0, y0) cartesian coordinates (in proj) of the upper left
            or lower left corner, depending on the sign of dy
        pixel_ref : str
            either 'center' or 'corner' (default: 'center'). Tells
            the Grid object where the (x0, y0) is located in the grid point.
            If ``pixel_ref`` is set to 'corner' and dy < 0, the ``x0y0``
            kwarg specifies the **grid point's upper left** corner
            coordinates.  Equivalently, if dy > 0, x0y0 specifies the
            **grid point's lower left** coordinate.
        corner : (float, float)
            DEPRECATED in favor of ``x0y0``
            (x0, y0) cartesian coordinates (in proj) of the upper left
            or lower left corner, depending on the sign of dy
        ul_corner : (float, float)
            DEPRECATED in favor of ``x0y0``
            (x0, y0) cartesian coordinates (in proj) of the upper left corner
        ll_corner : (float, float)
            DEPRECATED in favor of ``x0y0``
            (x0, y0) cartesian coordinates (in proj) of the lower left corner

        Examples
        --------
        >>> g = Grid(nxny=(3, 2), dxdy=(1, 1), x0y0=(0, 0), proj=wgs84)
        >>> lon, lat = g.ll_coordinates
        >>> lon
        array([[ 0.,  1.,  2.],
               [ 0.,  1.,  2.]])
        >>> lat
        array([[ 0.,  0.,  0.],
               [ 1.,  1.,  1.]])
        >>> lon, lat = g.corner_grid.ll_coordinates
        >>> lon
        array([[-0.5,  0.5,  1.5],
               [-0.5,  0.5,  1.5]])
        >>> lat
        array([[-0.5, -0.5, -0.5],
               [ 0.5,  0.5,  0.5]])
        >>> g.corner_grid == g.center_grid  # the two reprs are equivalent
        True
        """

        # Check for coordinate system
        proj = check_crs(proj)
        if proj is None:
            raise ValueError('proj must be of type pyproj.Proj')
        self._proj = proj

        # deprecations
        if corner is not None:
            warnings.warn('The `corner` kwarg is deprecated: '
                          'use `x0y0` instead.', DeprecationWarning)
            x0y0 = corner
        if ul_corner is not None:
            warnings.warn('The `ul_corner` kwarg is deprecated: '
                          'use `x0y0` instead.', DeprecationWarning)
            if dxdy[1] > 0.:
                raise ValueError('dxdy and input params not compatible')
            x0y0 = ul_corner
        if ll_corner is not None:
            warnings.warn('The `ll_corner` kwarg is deprecated: '
                          'use `x0y0` instead.', DeprecationWarning)
            if dxdy[1] < 0.:
                raise ValueError('dxdy and input params not compatible')
            x0y0 = ll_corner

        # Check for shortcut
        if dxdy[1] < 0.:
            ul_corner = x0y0
        else:
            ll_corner = x0y0

        # Initialise the rest
        self._check_input(nxny=nxny, dxdy=dxdy,
                          ul_corner=ul_corner,
                          ll_corner=ll_corner,
                          pixel_ref=pixel_ref)

    def _check_input(self, **kwargs):
        """See which parameter combination we have and set everything."""

        combi_a = ['nxny', 'dxdy', 'ul_corner']
        combi_b = ['nxny', 'dxdy', 'll_corner']
        if all(kwargs[k] is not None for k in combi_a):
            nx, ny = kwargs['nxny']
            dx, dy = kwargs['dxdy']
            x0, y0 = kwargs['ul_corner']
            if (dx <= 0.) or (dy >= 0.):
                raise ValueError('dxdy and input params not compatible')
            origin = 'upper-left'
        elif all(kwargs[k] is not None for k in combi_b):
            nx, ny = kwargs['nxny']
            dx, dy = kwargs['dxdy']
            x0, y0 = kwargs['ll_corner']
            if (dx <= 0.) or (dy <= 0.):
                raise ValueError('dxdy and input params not compatible')
            origin = 'lower-left'
        else:
            raise ValueError('Input params not compatible')

        self._nx = int(nx)
        self._ny = int(ny)
        if (self._nx <= 0) or (self._ny <= 0):
            raise ValueError('nxny not valid')
        self._dx = float(dx)
        self._dy = float(dy)
        self._x0 = float(x0)
        self._y0 = float(y0)
        self._origin = origin

        # Check for pixel ref
        self._pixel_ref = kwargs['pixel_ref'].lower()
        if self._pixel_ref not in ['corner', 'center']:
            raise ValueError('pixel_ref not recognized')

    def __eq__(self, other):
        """Two grids are considered equal when their defining coordinates
        and projection are equal.

        Note: equality also means floating point equality, with all the
        problems that come with it.

        (independent of the grid's cornered or centered representation.)
        """

        # Attributes defining the instance
        ckeys = ['x0', 'y0', 'nx', 'ny', 'dx', 'dy', 'origin']

        a = dict((k, getattr(self.corner_grid, k)) for k in ckeys)
        b = dict((k, getattr(other.corner_grid, k)) for k in ckeys)
        p1 = self.corner_grid.proj
        p2 = other.corner_grid.proj
        return (a == b) and proj_is_same(p1, p2)

    def __repr__(self):
        srs = '+'.join(sorted(self.proj.srs.split('+'))).strip()
        summary = ['<salem.Grid>']
        summary += ['  proj: ' + srs]
        summary += ['  pixel_ref: ' + self.pixel_ref]
        summary += ['  origin: ' + str(self.origin)]
        summary += ['  (nx, ny): (' + str(self.nx) + ', ' + str(self.ny) + ')']
        summary += ['  (dx, dy): (' + str(self.dx) + ', ' + str(self.dy) + ')']
        summary += ['  (x0, y0): (' + str(self.x0) + ', ' + str(self.y0) + ')']
        return '\n'.join(summary) + '\n'

    @property
    def proj(self):
        """``pyproj.Proj`` instance defining the grid's map projection."""
        return self._proj

    @property
    def nx(self):
        """number of grid points in the x direction."""
        return self._nx

    @property
    def ny(self):
        """number of grid points in the y direction."""
        return self._ny

    @property
    def dx(self):
        """x grid spacing (always positive)."""
        return self._dx

    @property
    def dy(self):
        """y grid spacing (positive if ll_corner, negative if ul_corner)."""
        return self._dy

    @property
    def x0(self):
        """X reference point in projection coordinates."""
        return self._x0

    @property
    def y0(self):
        """Y reference point in projection coordinates."""
        return self._y0

    @property
    def origin(self):
        """``'upper-left'`` or ``'lower-left'``."""
        return self._origin

    @property
    def pixel_ref(self):
        """if coordinates are at the ``'center'`` or ``'corner'`` of the grid.
        """
        return self._pixel_ref

    @lazy_property
    def center_grid(self):
        """``salem.Grid`` instance representing the grid in center coordinates.
        """

        if self.pixel_ref == 'center':
            return self
        else:
            # shift the grid
            x0y0 = ((self.x0 + self.dx / 2.), (self.y0 + self.dy / 2.))
            args = dict(nxny=(self.nx, self.ny), dxdy=(self.dx, self.dy),
                        proj=self.proj, pixel_ref='center', x0y0=x0y0)
            return Grid(**args)

    @lazy_property
    def corner_grid(self):
        """``salem.Grid`` instance representing the grid in corner coordinates.
        """

        if self.pixel_ref == 'corner':
            return self
        else:
            # shift the grid
            x0y0 = ((self.x0 - self.dx / 2.), (self.y0 - self.dy / 2.))
            args = dict(nxny=(self.nx, self.ny), dxdy=(self.dx, self.dy),
                        proj=self.proj, pixel_ref='corner', x0y0=x0y0)
            return Grid(**args)

    @property
    def ij_coordinates(self):
        """Tuple of i, j coordinates of the grid points.

        (dependent of the grid's cornered or centered representation.)
        """

        x = np.arange(self.nx)
        y = np.arange(self.ny)
        return np.meshgrid(x, y)

    @property
    def x_coord(self):
        """x coordinates of the grid points (1D, no mesh)"""

        return self.x0 + np.arange(self.nx) * self.dx

    @property
    def y_coord(self):
        """y coordinates of the grid points (1D, no mesh)"""

        return self.y0 + np.arange(self.ny) * self.dy

    @property
    def xy_coordinates(self):
        """Tuple of x, y coordinates of the grid points.

        (dependent of the grid's cornered or centered representation.)
        """

        return np.meshgrid(self.x_coord, self.y_coord)

    @lazy_property
    def ll_coordinates(self):
        """Tuple of longitudes, latitudes of the grid points.

        (dependent of the grid's cornered or centered representation.)
        """

        x, y = self.xy_coordinates
        proj_out = check_crs('EPSG:4326')

        return transform_proj(self.proj, proj_out, x, y)

    @property
    def xstagg_xy_coordinates(self):
        """Tuple of x, y coordinates of the X staggered grid.

        (independent of the grid's cornered or centered representation.)
        """

        x_s = self.corner_grid.x0 + np.arange(self.nx+1) * self.dx
        y = self.center_grid.y0 + np.arange(self.ny) * self.dy
        return np.meshgrid(x_s, y)

    @property
    def ystagg_xy_coordinates(self):
        """Tuple of x, y coordinates of the Y staggered grid.

        (independent of the grid's cornered or centered representation.)
        """

        x = self.center_grid.x0 + np.arange(self.nx) * self.dx
        y_s = self.corner_grid.y0 + np.arange(self.ny+1) * self.dy
        return np.meshgrid(x, y_s)

    @lazy_property
    def xstagg_ll_coordinates(self):
        """Tuple of longitudes, latitudes of the X staggered grid.

        (independent of the grid's cornered or centered representation.)
        """

        x, y = self.xstagg_xy_coordinates
        proj_out = check_crs('EPSG:4326')
        return transform_proj(self.proj, proj_out, x, y)

    @lazy_property
    def ystagg_ll_coordinates(self):
        """Tuple of longitudes, latitudes of the Y staggered grid.

        (independent of the grid's cornered or centered representation.)
        """

        x, y = self.ystagg_xy_coordinates
        proj_out = check_crs('EPSG:4326')
        return transform_proj(self.proj, proj_out, x, y)

    @lazy_property
    def pixcorner_ll_coordinates(self):
        """Tuple of longitudes, latitudes (dims: ny+1, nx+1) at the corners of
        the grid.

        Useful for graphics.Map essentially

        (independant of the grid's cornered or centered representation.)
        """

        x = self.corner_grid.x0 + np.arange(self.nx+1) * self.dx
        y = self.corner_grid.y0 + np.arange(self.ny+1) * self.dy
        x, y = np.meshgrid(x, y)
        proj_out = check_crs('EPSG:4326')
        return transform_proj(self.proj, proj_out, x, y)

    @lazy_property
    def extent(self):
        """[left, right, bottom, top] boundaries of the grid in the grid's
        projection.

        The boundaries are the pixels leftmost, rightmost, lowermost and
        uppermost corners, meaning that they are independent from the grid's
        representation.
        """

        x = np.array([0, self.nx]) * self.dx + self.corner_grid.x0
        ypoint = [0, self.ny] if self.origin == 'lower-left' else [self.ny, 0]
        y = np.array(ypoint) * self.dy + self.corner_grid.y0

        return [x[0], x[1], y[0], y[1]]

    def almost_equal(self, other, rtol=1e-05, atol=1e-08):
        """A less strict comparison between grids.

        Two grids are considered equal when their defining coordinates
        and projection are equal.
        grid1 == grid2 uses floating point equality, which is very strict; here
        we uses numpy's is close instead.

        (independent of the grid's cornered or centered representation.)
        """

        # float attributes defining the instance
        fkeys = ['x0', 'y0', 'dx', 'dy']
        # unambiguous attributes
        ckeys = ['nx', 'ny', 'origin']

        ok = True
        for k in fkeys:
            ok = ok and np.isclose(getattr(self.corner_grid, k),
                                   getattr(other.corner_grid, k),
                                   rtol=rtol, atol=atol)
        for k in ckeys:
            _ok = getattr(self.corner_grid, k) == getattr(other.corner_grid, k)
            ok = ok and _ok
        p1 = self.corner_grid.proj
        p2 = other.corner_grid.proj
        return ok and proj_is_same(p1, p2)

    def extent_in_crs(self, crs=wgs84):
        """Get the extent of the grid in a desired crs.

        Parameters
        ----------
        crs : crs
            the target coordinate reference system.

        Returns
        -------
        [left, right, bottom, top] boundaries of the grid.
        """

        # this is not so trivial
        # for optimisation we will transform the boundaries only
        poly = self.extent_as_polygon(crs=crs)
        _i, _j = poly.exterior.xy
        return [np.min(_i), np.max(_i), np.min(_j), np.max(_j)]

    def extent_as_polygon(self, crs=wgs84):
        """Get the extent of the grid in a shapely.Polygon and desired crs.

        Parameters
        ----------
        crs : crs
            the target coordinate reference system.

        Returns
        -------
        [left, right, bottom, top] boundaries of the grid.
        """
        from shapely.geometry import Polygon

        # this is not so trivial
        # for optimisation we will transform the boundaries only
        _i = np.hstack([np.arange(self.nx+1),
                        np.ones(self.ny+1)*self.nx,
                        np.arange(self.nx+1)[::-1],
                        np.zeros(self.ny+1)]).flatten()
        _j = np.hstack([np.zeros(self.nx+1),
                        np.arange(self.ny+1),
                        np.ones(self.nx+1)*self.ny,
                        np.arange(self.ny+1)[::-1]]).flatten()
        _i, _j = self.corner_grid.ij_to_crs(_i, _j, crs=crs)
        return Polygon(zip(_i, _j))

    def regrid(self, nx=None, ny=None, factor=1):
        """Make a copy of the grid with an updated spatial resolution.

        The keyword parameters are mutually exclusive, because the x/y ratio
        of the grid has to be preserved.

        Parameters
        ----------
        nx : int
            the new number of x pixels
        nx : int
            the new number of y pixels
        factor : int
            multiplication factor (factor=3 will generate a grid with
            a spatial resolution 3 times finer)

        Returns
        -------
        a new Grid object.
        """

        if nx is not None:
            factor = nx / self.nx
        if ny is not None:
            factor = ny / self.ny

        nx = self.nx * factor
        ny = self.ny * factor
        dx = self.dx / factor
        dy = self.dy / factor

        x0 = self.corner_grid.x0
        y0 = self.corner_grid.y0
        args = dict(nxny=(nx, ny), dxdy=(dx, dy), x0y0=(x0, y0),
                    proj=self.proj, pixel_ref='corner')
        g = Grid(**args)
        if self.pixel_ref == 'center':
            g = g.center_grid
        return g

    def ij_to_crs(self, i, j, crs=None, nearest=False):
        """Converts local i, j to cartesian coordinates in a specified crs

        Parameters
        ----------
        i : array of floats
            the grid coordinates of the point(s) you want to convert
        j : array of floats
            the grid coordinates of the point(s) you want to convert
        crs: crs
             the target crs (default: self.proj)
        nearest: bool
             (for Grid crs only) convert to the nearest grid point

        Returns
        -------
        (x, y) coordinates of the points in the specified crs.
        """

        # Default
        if crs is None:
            crs = self.proj

        # Convert i, j to x, y
        try:
            x = i * self.dx + self.x0
            y = j * self.dy + self.y0
        except TypeError:
            x = np.asarray(i) * self.dx + self.x0
            y = np.asarray(j) * self.dy + self.y0

        # Convert x, y to crs
        _crs = check_crs(crs, raise_on_error=True)
        if isinstance(_crs, pyproj.Proj):
            ret = transform_proj(self.proj, _crs, x, y)
        elif isinstance(_crs, Grid):
            ret = _crs.transform(x, y, crs=self.proj, nearest=nearest)
        return ret

    def transform(self, x, y, z=None, crs=wgs84, nearest=False, maskout=False):
        """Converts any coordinates into the local grid.

        Parameters
        ----------
        x : ndarray
            the grid coordinates of the point(s) you want to convert
        y : ndarray
            the grid coordinates of the point(s) you want to convert
        z : None
            ignored (but necessary since some shapes have a z dimension)
        crs : crs
            reference system of x, y. Could be a pyproj.Proj instance or a
            Grid instance. In the latter case (x, y) are actually (i, j).
            (Default: lonlat in wgs84).
        nearest : bool
            set to True if you wish to return the closest i, j coordinates
            instead of subpixel coords.
        maskout : bool
            set to true if you want to mask out the transformed
            coordinates that are not within the grid.

        Returns
        -------
        (i, j) coordinates of the points in the local grid.
        """

        x, y = np.ma.array(x), np.ma.array(y)

        # First to local proj
        _crs = check_crs(crs, raise_on_error=True)
        if isinstance(_crs, pyproj.Proj):
            x, y = transform_proj(_crs, self.proj, x, y)
        elif isinstance(_crs, Grid):
            x, y = _crs.ij_to_crs(x, y, crs=self.proj)

        # Then to local grid
        x = (x - self.x0) / self.dx
        y = (y - self.y0) / self.dy

        # See if we need to round
        if nearest:
            f = np.rint if self.pixel_ref == 'center' else np.floor
            x = f(x).astype(int)
            y = f(y).astype(int)

        # Mask?
        if maskout:
            if self.pixel_ref == 'center':
                mask = ~((x >= -0.5) & (x < self.nx-0.5) &
                         (y >= -0.5) & (y < self.ny-0.5))
            else:
                mask = ~((x >= 0) & (x < self.nx) &
                         (y >= 0) & (y < self.ny))
            x = np.ma.array(x, mask=mask)
            y = np.ma.array(y, mask=mask)

        return x, y

    def grid_lookup(self, other):
        """Performs forward transformation of any other grid into self.

        The principle of forward transform is to obtain, for each grid point of
        ``self`` , all the indices of ``other`` that are located into the
        given grid point. This transformation makes sense ONLY if ``other`` has
        a higher resolution than the object grid. If ``other`` has a similar
        or coarser resolution than ``self`` , choose the more general
        (and much faster) :py:meth:`Grid.map_gridded_data` method.

        Parameters
        ----------
        other : salem.Grid
            the grid that needs to be transformed into self

        Returns
        -------
        a dict: each key (j, i) contains an array of shape (n, 2) where n is
        the number of ``other`` 's grid points found within the grid point
        (j, i)
        """

        # Input checks
        other = check_crs(other)
        if not isinstance(other, Grid):
            raise ValueError('`other` should be a Grid instance')

        # Transform the other grid into the local grid (forward transform)
        # Work in center grid cause that's what we need
        i, j = other.center_grid.ij_coordinates
        i, j = i.flatten(), j.flatten()
        oi, oj = self.center_grid.transform(i, j, crs=other.center_grid,
                                            nearest=True, maskout=True)
        # keep only valid values
        oi, oj, i, j = oi[~oi.mask], oj[~oi.mask], i[~oi.mask], j[~oi.mask]

        out_inds = oi.flatten() + self.nx * oj.flatten()

        # find the links
        ris = np.digitize(out_inds, bins=np.arange(self.nx*self.ny+1))

        # some optim based on the fact that ris has many duplicates
        sort_idx = np.argsort(ris)
        unq_items, unq_count = np.unique(ris[sort_idx], return_counts=True)
        unq_idx = np.split(sort_idx, np.cumsum(unq_count))

        # lets go
        out = dict()
        for idx, ri in zip(unq_idx, unq_items):
            ij = divmod(ri-1, self.nx)
            out[ij] = np.stack((j[idx], i[idx]), axis=1)
        return out

    def lookup_transform(self, data, grid=None, method=np.mean, lut=None,
                         return_lut=False):
        """Performs the forward transformation of gridded data into self.

        This method is suitable when the data grid is of higher resolution
        than ``self``. ``lookup_transform`` performs aggregation of data
        according to a user given rule (e.g. ``np.mean``, ``len``, ``np.std``),
        applied to all grid points found below a grid point in ``self``.


        See also :py:meth:`Grid.grid_lookup` and examples in the docs

        Parameters
        ----------
        data : ndarray
            an ndarray of dimensions 2, 3, or 4, the two last ones being y, x.
        grid : Grid
            a Grid instance matching the data
        method : function, default: np.mean
            the aggregation method. Possibilities: np.std, np.median, np.sum,
            and more. Use ``len`` to count the number of grid points!
        lut : ndarray, optional
            computing the lookup table can be expensive. If you have several
            operations to do with the same grid, set ``lut`` to an existing
            table obtained from a previous call to  :py:meth:`Grid.grid_lookup`
        return_lut : bool, optional
            set to True if you want to return the lookup table for later use.
            in this case, returns a tuple

        Returns
        -------
        An aggregated ndarray of the data, in ``self`` coordinates.
        If ``return_lut==True``, also return the lookup table
        """

        # Input checks
        if grid is None:
            grid = check_crs(data)  # xarray
        if not isinstance(grid, Grid):
            raise ValueError('grid should be a Grid instance')
        if hasattr(data, 'values'):
            data = data.values  # xarray

        # dimensional check
        in_shape = data.shape
        ndims = len(in_shape)
        if (ndims < 2) or (ndims > 4):
            raise ValueError('data dimension not accepted')
        if (in_shape[-1] != grid.nx) or (in_shape[-2] != grid.ny):
            raise ValueError('data dimension not compatible')

        if lut is None:
            lut = self.grid_lookup(grid)

        # Prepare the output
        out_shape = list(in_shape)
        out_shape[-2:] = [self.ny, self.nx]

        if data.dtype.kind == 'i':
            out_data = np.zeros(out_shape, dtype=float) * np.NaN
        else:
            out_data = np.zeros(out_shape, dtype=data.dtype) * np.NaN

        def _2d_trafo(ind, outd):
            for ji, l in lut.items():
                outd[ji] = method(ind[l[:, 0], l[:, 1]])
            return outd

        if ndims == 2:
            _2d_trafo(data, out_data)
        if ndims == 3:
            for dimi, cdata in enumerate(data):
                out_data[dimi, ...] = _2d_trafo(cdata, out_data[dimi, ...])
        if ndims == 4:
            for dimj, cdata in enumerate(data):
                for dimi, ccdata in enumerate(cdata):
                    tmp = _2d_trafo(ccdata, out_data[dimj, dimi, ...])
                    out_data[dimj, dimi, ...] = tmp

        # prepare output
        if method is len:
            out_data[~np.isfinite(out_data)] = 0
            out_data = out_data.astype(int)
        else:
            out_data = np.ma.masked_invalid(out_data)

        if return_lut:
            return out_data, lut
        else:
            return out_data

    def map_gridded_data(self, data, grid=None, interp='nearest',
                         ks=3, out=None):
        """Reprojects any structured data onto the local grid.

        The z and time dimensions of the data (if provided) are conserved, but
        the projected data will have the x, y dimensions of the local grid.

        Currently, nearest neighbor, linear, and spline interpolation are
        available. The dtype of the input data is guaranteed to be conserved,
        except for int which will be converted to floats if non nearest
        neighbor interpolation is asked.

        Parameters
        ----------
        data : ndarray
            an ndarray of dimensions 2, 3, or 4, the two last ones being y, x.
        grid : Grid
            a Grid instance matching the data
        interp : str
            'nearest' (default), 'linear', or 'spline'
        ks : int
            Degree of the bivariate spline. Default is 3.
        missing : int
            integer value to attribute to invalid data (for integer data
            only, floats invalids are forced to NaNs)
        out : ndarray
            output array to fill instead of creating a new one (useful for
            overwriting stuffs)

        Returns
        -------
        A projected ndarray of the data, in ``self`` coordinates.
        """

        if grid is None:
            try:
                grid = data.salem.grid  # try xarray
            except AttributeError:
                pass

        # Input checks
        if not isinstance(grid, Grid):
            raise ValueError('grid should be a Grid instance')

        try:  # in case someone gave an xarray dataarray
            data = data.values
        except AttributeError:
            pass

        try:  # in case someone gave a masked array (won't work with scipy)
            data = data.filled(np.nan)
        except AttributeError:
            pass

        in_shape = data.shape
        ndims = len(in_shape)
        if (ndims < 2) or (ndims > 4):
            raise ValueError('data dimension not accepted')
        if (in_shape[-1] != grid.nx) or (in_shape[-2] != grid.ny):
            raise ValueError('data dimension not compatible')

        interp = interp.lower()

        use_nn = False
        if interp == 'nearest':
            use_nn = True

        # Transform the local grid into the input grid (backwards transform)
        # Work in center grid cause that's what we need
        # TODO: this stage could be optimized when many variables need transfo
        i, j = self.center_grid.ij_coordinates
        oi, oj = grid.center_grid.transform(i, j, crs=self.center_grid,
                                            nearest=use_nn, maskout=False)
        pv = np.nonzero((oi >= 0) & (oi < grid.nx) &
                        (oj >= 0) & (oj < grid.ny))

        # Prepare the output
        if out is not None:
            out_data = np.ma.asarray(out)
        else:
            out_shape = list(in_shape)
            out_shape[-2:] = [self.ny, self.nx]
            if (data.dtype.kind == 'i') and (interp == 'nearest'):
                # We dont do integer arithmetics other than nearest
                out_data = np.ma.masked_all(out_shape, dtype=data.dtype)
            elif data.dtype.kind == 'i':
                out_data = np.ma.masked_all(out_shape, dtype=float)
            else:
                out_data = np.ma.masked_all(out_shape, dtype=data.dtype)

        # Spare us the trouble
        if len(pv[0]) == 0:
            return out_data

        i, j, oi, oj = i[pv], j[pv], oi[pv], oj[pv]

        # Interpolate
        if interp == 'nearest':
            if out is not None:
                if ndims > 2:
                    raise ValueError('Need 2D for now.')
                vok = np.isfinite(data[oj, oi])
                out_data[j[vok], i[vok]] = data[oj[vok], oi[vok]]
            else:
                out_data[..., j, i] = data[..., oj, oi]
        elif interp == 'linear':
            points = (np.arange(grid.ny), np.arange(grid.nx))
            if ndims == 2:
                f = RegularGridInterpolator(points, data, bounds_error=False)
                if out is not None:
                    tmp = f((oj, oi))
                    vok = np.isfinite(tmp)
                    out_data[j[vok], i[vok]] = tmp[vok]
                else:
                    out_data[j, i] = f((oj, oi))
            if ndims == 3:
                for dimi, cdata in enumerate(data):
                    f = RegularGridInterpolator(points, cdata,
                                                bounds_error=False)
                    if out is not None:
                        tmp = f((oj, oi))
                        vok = np.isfinite(tmp)
                        out_data[dimi, j[vok], i[vok]] = tmp[vok]
                    else:
                        out_data[dimi, j, i] = f((oj, oi))
            if ndims == 4:
                for dimj, cdata in enumerate(data):
                    for dimi, ccdata in enumerate(cdata):
                        f = RegularGridInterpolator(points, ccdata,
                                                    bounds_error=False)
                        if out is not None:
                            tmp = f((oj, oi))
                            vok = np.isfinite(tmp)
                            out_data[dimj, dimi, j[vok], i[vok]] = tmp[vok]
                        else:
                            out_data[dimj, dimi, j, i] = f((oj, oi))
        elif interp == 'spline':
            px, py = np.arange(grid.ny), np.arange(grid.nx)
            if ndims == 2:
                f = RectBivariateSpline(px, py, data, kx=ks, ky=ks)
                if out is not None:
                    tmp = f(oj, oi, grid=False)
                    vok = np.isfinite(tmp)
                    out_data[j[vok], i[vok]] = tmp[vok]
                else:
                    out_data[j, i] = f(oj, oi, grid=False)
            if ndims == 3:
                for dimi, cdata in enumerate(data):
                    f = RectBivariateSpline(px, py, cdata, kx=ks, ky=ks)
                    if out is not None:
                        tmp = f(oj, oi, grid=False)
                        vok = np.isfinite(tmp)
                        out_data[dimi, j[vok], i[vok]] = tmp[vok]
                    else:
                        out_data[dimi, j, i] = f(oj, oi, grid=False)
            if ndims == 4:
                for dimj, cdata in enumerate(data):
                    for dimi, ccdata in enumerate(cdata):
                        f = RectBivariateSpline(px, py, ccdata, kx=ks, ky=ks)
                        if out is not None:
                            tmp = f(oj, oi, grid=False)
                            vok = np.isfinite(tmp)
                            out_data[dimj, dimi, j[vok], i[vok]] = tmp[vok]
                        else:
                            out_data[dimj, dimi, j, i] = f(oj, oi, grid=False)
        else:
            msg = 'interpolation not understood: {}'.format(interp)
            raise ValueError(msg)

        # we have to catch a warning for an unexplained reason
        with warnings.catch_warnings():
            mess = "invalid value encountered in isfinite"
            warnings.filterwarnings("ignore", message=mess)
            out_data = np.ma.masked_invalid(out_data)
        return out_data

    def region_of_interest(self, shape=None, geometry=None, grid=None,
                           corners=None, crs=wgs84, roi=None,
                           all_touched=False):
        """Computes a region of interest (ROI).

        A ROI is simply a mask of the same size as the grid.

        Parameters
        ----------
        shape : str
            path to a shapefile
        geometry : geometry
            a shapely geometry (don't forget the crs keyword)
        grid : Grid
            a Grid object which extent will form the ROI
        corners : tuple
            a ((x0, y0), (x1, y1)) tuple of the corners of the square
            to subset the dataset to  (don't forget the crs keyword)
        crs : crs, default wgs84
            coordinate reference system of the geometry and corners
        roi : ndarray
            add the new region_of_interest to a previous one (useful for
            multiple geometries for example)
        all_touched : boolean
            pass-through argument for rasterio.features.rasterize, indicating
            that all grid cells which are  clipped by the shapefile defining
            the region of interest should be included (default=False)
        """

        # Initial mask
        if roi is not None:
            mask = np.array(roi, dtype=np.int16)
        else:
            mask = np.zeros((self.ny, self.nx), dtype=np.int16)

        # Collect keyword arguments, overriding anything the user
        # inadvertently added
        rasterize_kws = dict(out=mask, all_touched=all_touched)

        # Several cases
        if shape is not None:
            import pandas as pd
            inplace = False
            if not isinstance(shape, pd.DataFrame):
                from salem.sio import read_shapefile
                shape = read_shapefile(shape)
                inplace = True
            # corner grid is needed for rasterio
            shape = transform_geopandas(shape, to_crs=self.corner_grid,
                                        inplace=inplace)
            import rasterio
            from rasterio.features import rasterize
            with rasterio.Env():
                mask = rasterize(shape.geometry, **rasterize_kws)
        if geometry is not None:
            import rasterio
            from rasterio.features import rasterize
            # corner grid is needed for rasterio
            geom = transform_geometry(geometry, crs=crs,
                                      to_crs=self.corner_grid)
            with rasterio.Env():
                mask = rasterize(np.atleast_1d(geom), **rasterize_kws)
        if grid is not None:
            _tmp = np.ones((grid.ny, grid.nx), dtype=np.int16)
            mask = self.map_gridded_data(_tmp, grid, out=mask).filled(0)
        if corners is not None:
            cgrid = self.center_grid
            xy0, xy1 = corners
            x0, y0 = cgrid.transform(*xy0, crs=crs, nearest=True)
            x1, y1 = cgrid.transform(*xy1, crs=crs, nearest=True)
            mask[np.min([y0, y1]):np.max([y0, y1]) + 1,
            np.min([x0, x1]):np.max([x0, x1]) + 1] = 1

        return mask

    def to_dict(self):
        """Serialize this grid to a dictionary.

        Returns
        -------
        a grid dictionary

        See Also
        --------
        from_dict : create a Grid from a dict
        """
        srs = self.proj.srs
        return dict(proj=self.proj.srs, x0y0=(self.x0, self.y0),
                    nxny=(self.nx, self.ny), dxdy=(self.dx, self.dy),
                    pixel_ref=self.pixel_ref)

    @classmethod
    def from_dict(self, d):
        """Create a Grid from a dictionary

        Parameters
        ----------
        d : dict, required
            the dict with the necessary information

        Returns
        -------
        a salem.Grid instance

        See Also
        --------
        to_dict : create a dict from a Grid
        """
        return Grid(**d)

    def to_json(self, fpath):
        """Serialize this grid to a json file.

        Parameters
        ----------
        fpath : str, required
            the path to the file to create

        See Also
        --------
        from_json : read a json file
        """
        import json
        with open(fpath, 'w') as fp:
            json.dump(self.to_dict(), fp)

    @classmethod
    def from_json(self, fpath):
        """Create a Grid from a json file

        Parameters
        ----------
        fpath : str, required
            the path to the file to open

        Returns
        -------
        a salem.Grid instance

        See Also
        --------
        to_json : create a json file
        """
        import json
        with open(fpath, 'r') as fp:
            d = json.load(fp)
        return Grid.from_dict(d)

    def to_dataset(self):
        """Creates an empty dataset based on the Grid's geolocalisation.

        Returns
        -------
        An xarray.Dataset object ready to be filled with data
        """
        import xarray as xr
        ds = xr.Dataset(coords={'x': (['x', ], self.center_grid.x_coord),
                                'y': (['y', ], self.center_grid.y_coord)}
                        )
        ds.attrs['pyproj_srs'] = self.proj.srs
        return ds

    def to_geometry(self, to_crs=None):
        """Makes a geometrical representation of the grid (e.g. for drawing).

        This can come also handy when doing shape-to-raster operations.

        TODO: currently returns one polygon for each grid point, but this
        could do more.

        Returns
        -------
        a geopandas.GeoDataFrame
        """
        from geopandas import GeoDataFrame
        from shapely.geometry import Polygon
        out = GeoDataFrame()
        geoms = []
        ii = []
        jj = []
        xx = self.corner_grid.x0 + np.arange(self.nx+1) * self.dx
        yy = self.corner_grid.y0 + np.arange(self.ny+1) * self.dy
        for j, (y0, y1) in enumerate(zip(yy[:-1], yy[1:])):
            for i, (x0, x1) in enumerate(zip(xx[:-1], xx[1:])):
                coords = [(x0, y0), (x1, y0), (x1, y1), (x0, y1), (x0, y0)]
                geoms.append(Polygon(coords))
                jj.append(j)
                ii.append(i)
        out['j'] = jj
        out['i'] = ii
        out['geometry'] = geoms
        out.crs = self.proj.srs

        if check_crs(to_crs):
            transform_geopandas(out, to_crs=to_crs, inplace=True)
        return out


def proj_is_same(p1, p2):
    """Checks is two pyproj projections are equal.

    See https://github.com/jswhit/pyproj/issues/15#issuecomment-208862786

    Parameters
    ----------
    p1 : pyproj.Proj
        first projection
    p2 : pyproj.Proj
        second projection
    """
    if has_gdal:
        # this is more robust, but gdal is a pain
        s1 = osr.SpatialReference()
        s1.ImportFromProj4(p1.srs)
        s2 = osr.SpatialReference()
        s2.ImportFromProj4(p2.srs)
        return s1.IsSame(s2) == 1  # IsSame returns 1 or 0
    else:
        # at least we can try to sort it
        p1 = '+'.join(sorted(p1.srs.split('+')))
        p2 = '+'.join(sorted(p2.srs.split('+')))
        return p1 == p2


def _transform_internal(p1, p2, x, y, **kwargs):
    if hasattr(pyproj, 'Transformer'):
        trf = pyproj.Transformer.from_proj(p1, p2, **kwargs)
        return trf.transform(x, y)
    else:
        return pyproj.transform(p1, p2, x, y, **kwargs)


def transform_proj(p1, p2, x, y, nocopy=False):
    """Wrapper around the pyproj.transform function.

    Transform points between two coordinate systems defined by the Proj
    instances p1 and p2.

    When two projections are equal, this function avoids quite a bunch of
    useless calculations. See https://github.com/jswhit/pyproj/issues/15

    Parameters
    ----------
    p1 : pyproj.Proj
        projection associated to x and y
    p2 : pyproj.Proj
        projection into which x, y must be transformed
    x : ndarray
        eastings
    y : ndarray
        northings
    nocopy : bool
        in case the two projections are equal, you can use nocopy if you wish
    """

    try:
        # This always makes a copy, even if projections are equivalent
        return _transform_internal(p1, p2, x, y, always_xy=True)
    except TypeError:
        if proj_is_same(p1, p2):
            if nocopy:
                return x, y
            else:
                return copy.deepcopy(x), copy.deepcopy(y)

        return _transform_internal(p1, p2, x, y)


def transform_geometry(geom, crs=wgs84, to_crs=wgs84):
    """Reprojects a shapely geometry.

    Parameters
    ----------
    geom : shapely geometry
        the geometry to transform
    crs : crs
        the geometry's crs
    to_crs : crs
        the crs into which the geometry must be transformed

    Returns
    -------
    A reprojected geometry
    """

    from_crs = check_crs(crs)
    to_crs = check_crs(to_crs)

    if isinstance(to_crs, pyproj.Proj) and isinstance(from_crs, pyproj.Proj):
        project = partial(transform_proj, from_crs, to_crs)
    elif isinstance(to_crs, Grid):
        project = partial(to_crs.transform, crs=from_crs)
    elif isinstance(from_crs, Grid):
        project = partial(from_crs.ij_to_crs, crs=to_crs)
    else:
        raise NotImplementedError()

    from shapely.ops import transform
    return transform(project, geom)


def transform_geopandas(gdf, from_crs=None, to_crs=wgs84, inplace=False):
    """Reprojects a geopandas dataframe.

    Parameters
    ----------
    gdf : geopandas.DataFrame
        the dataframe to transform (must have a crs attribute)
    from_crs : crs
        if gdf has no crs attribute (happens when the crs is a salem grid)
    to_crs : crs
        the crs into which the dataframe must be transformed
    inplace : bool
        the original dataframe will be overwritten (default: False)

    Returns
    -------
    A projected dataframe
    """
    from shapely.ops import transform
    import geopandas as gpd

    if from_crs is None:
        from_crs = check_crs(gdf.crs)
    else:
        from_crs = check_crs(from_crs)
    to_crs = check_crs(to_crs)

    if inplace:
        out = gdf
    else:
        out = gdf.copy()

    if isinstance(to_crs, pyproj.Proj) and isinstance(from_crs, pyproj.Proj):
        project = partial(transform_proj, from_crs, to_crs)
    elif isinstance(to_crs, Grid):
        project = partial(to_crs.transform, crs=from_crs)
    elif isinstance(from_crs, Grid):
        project = partial(from_crs.ij_to_crs, crs=to_crs)
    else:
        raise NotImplementedError()

    # Do the job and set the new attributes
    result = out.geometry.apply(lambda geom: transform(project, geom))
    result.__class__ = gpd.GeoSeries
    if isinstance(to_crs, pyproj.Proj):
        to_crs = to_crs.srs
    elif isinstance(to_crs, Grid):
        to_crs = None
    result.crs = to_crs
    out.geometry = result
    out.crs = to_crs
    out['min_x'] = [g.bounds[0] for g in out.geometry]
    out['max_x'] = [g.bounds[2] for g in out.geometry]
    out['min_y'] = [g.bounds[1] for g in out.geometry]
    out['max_y'] = [g.bounds[3] for g in out.geometry]
    return out


def proj_is_latlong(proj):
    """Shortcut function because of deprecation."""

    try:
        return proj.is_latlong()
    except AttributeError:
        return proj.crs.is_geographic


def proj_to_cartopy(proj):
    """Converts a pyproj.Proj to a cartopy.crs.Projection

    Parameters
    ----------
    proj: pyproj.Proj
        the projection to convert

    Returns
    -------
    a cartopy.crs.Projection object

    """

    import cartopy
    import cartopy.crs as ccrs

    proj = check_crs(proj)

    if proj_is_latlong(proj):
        return ccrs.PlateCarree()

    srs = proj.srs
    if has_gdal:
        # this is more robust, as srs could be anything (espg, etc.)
        from osgeo import osr
        s1 = osr.SpatialReference()
        s1.ImportFromProj4(proj.srs)
        if s1.ExportToProj4():
            srs = s1.ExportToProj4()

    km_proj = {'lon_0': 'central_longitude',
               'lat_0': 'central_latitude',
               'x_0': 'false_easting',
               'y_0': 'false_northing',
               'lat_ts': 'latitude_true_scale',
               'o_lon_p': 'central_rotated_longitude',
               'o_lat_p': 'pole_latitude',
               'k': 'scale_factor',
               'zone': 'zone',
               }
    km_globe = {'a': 'semimajor_axis',
                'b': 'semiminor_axis',
                }
    km_std = {'lat_1': 'lat_1',
              'lat_2': 'lat_2',
              }
    kw_proj = dict()
    kw_globe = dict()
    kw_std = dict()
    for s in srs.split('+'):
        s = s.split('=')
        if len(s) != 2:
            continue
        k = s[0].strip()
        v = s[1].strip()
        try:
            v = float(v)
        except:
            pass
        if k == 'proj':
            if v == 'tmerc':
                cl = ccrs.TransverseMercator
                kw_proj['approx'] = True
            if v == 'lcc':
                cl = ccrs.LambertConformal
            if v == 'merc':
                cl = ccrs.Mercator
            if v == 'utm':
                cl = ccrs.UTM
            if v == 'stere':
                cl = ccrs.Stereographic
            if v == 'ob_tran':
                cl = ccrs.RotatedPole
        if k in km_proj:
            if k == 'zone':
                v = int(v)
            kw_proj[km_proj[k]] = v
        if k in km_globe:
            kw_globe[km_globe[k]] = v
        if k in km_std:
            kw_std[km_std[k]] = v

    globe = None
    if kw_globe:
        globe = ccrs.Globe(ellipse='sphere', **kw_globe)
    if kw_std:
        kw_proj['standard_parallels'] = (kw_std['lat_1'], kw_std['lat_2'])

    # mercatoooor
    if cl.__name__ == 'Mercator':
        kw_proj.pop('false_easting', None)
        kw_proj.pop('false_northing', None)
        if LooseVersion(cartopy.__version__) < LooseVersion('0.15'):
            kw_proj.pop('latitude_true_scale', None)
    elif cl.__name__ == 'Stereographic':
        kw_proj.pop('scale_factor', None)
        if 'latitude_true_scale' in kw_proj:
            kw_proj['true_scale_latitude'] = kw_proj['latitude_true_scale']
            kw_proj.pop('latitude_true_scale', None)
    elif cl.__name__ == 'RotatedPole':
        if 'central_longitude' in kw_proj:
            kw_proj['pole_longitude'] = kw_proj['central_longitude'] - 180
            kw_proj.pop('central_longitude', None)
    else:
        kw_proj.pop('latitude_true_scale', None)

    try:
        return cl(globe=globe, **kw_proj)
    except TypeError:
        del kw_proj['approx']

    return cl(globe=globe, **kw_proj)


def mercator_grid(center_ll=None, extent=None, ny=600, nx=None,
                  origin='lower-left', transverse=True):
    """Local (transverse) mercator map centered on a specified point.

    Parameters
    ----------
    center_ll : (float, float)
        tuple of lon, lat coordinates where the map will be centered.
    extent : (float, float)
        tuple of eastings, northings giving the extent (in m) of the map
    ny : int
        number of y grid points wanted to cover the map (default: 600)
    nx : int
        number of x grid points wanted to cover the map (mutually exclusive
        with y)
    origin : str
        'lower-left' or 'upper-left'
    transverse : bool
        wether to use a transverse or regular mercator. Default should have
        been false, but for backwards compatibility reasons we keep it to True
    """

    # Make a local proj
    pname = 'tmerc' if transverse else 'merc'
    lon, lat = center_ll
    proj_params = dict(proj=pname, lat_0=0., lon_0=lon,
                       k=0.9996, x_0=0, y_0=0, datum='WGS84')
    projloc = pyproj.Proj(proj_params)

    # Define a spatial resolution
    xx = extent[0]
    yy = extent[1]
    if nx is None:
        nx = ny * xx / yy
    else:
        ny = nx * yy / xx

    nx = np.rint(nx)
    ny = np.rint(ny)

    e, n = transform_proj(wgs84, projloc, lon, lat)

    if origin== 'upper-left':
        corner = (-xx / 2. + e, yy / 2. + n)
        dxdy = (xx / nx, - yy / ny)
    else:
        corner = (-xx / 2. + e, -yy / 2. + n)
        dxdy = (xx / nx, yy / ny)

    return Grid(proj=projloc, x0y0=corner, nxny=(nx, ny), dxdy=dxdy,
                pixel_ref='corner')


def googlestatic_mercator_grid(center_ll=None, nx=640, ny=640, zoom=12, scale=1):
    """Mercator map centered on a specified point (google API conventions).

    Mostly useful for google maps.
    """

    # Number of pixels in an image with a zoom level of 0.
    google_pix = 256 * scale
    # The equatorial radius of the Earth assuming WGS-84 ellipsoid.
    google_earth_radius = 6378137.0

    # Make a local proj
    lon, lat = center_ll
    projloc = check_crs('epsg:3857')

    # The size of the image is multiplied by the scaling factor
    nx *= scale
    ny *= scale

    # Meter per pixel
    mpix = (2 * np.pi * google_earth_radius) / google_pix / (2**zoom)
    xx = nx * mpix
    yy = ny * mpix

    e, n = transform_proj(wgs84, projloc, lon, lat)
    corner = (-xx / 2. + e, yy / 2. + n)
    dxdy = (xx / nx, - yy / ny)

    return Grid(proj=projloc, x0y0=corner,
                nxny=(nx, ny), dxdy=dxdy,
                pixel_ref='corner')



"""
This module provides a GeoDataset interface and a few implementations for
e.g. netcdf, geotiff, WRF...

This is kept for backwards compatibility reasons, but ideally everything should
soon happen at the xarray level.
"""
from __future__ import division

# Builtins
import io
import os
import warnings
from six.moves.urllib.request import urlopen

# External libs
import pyproj
import numpy as np
import netCDF4
import pandas as pd
import xarray as xr

try:
    import rasterio
except ImportError:
    rasterio = None

# Locals
from salem import lazy_property
from salem import Grid
from salem import wgs84
from salem import utils, gis, wrftools, sio, check_crs

API_KEY = None


def _to_scalar(x):
    """If a list then scalar"""
    try:
        return x[0]
    except IndexError:
        return x


class GeoDataset(object):
    """Interface for georeferenced datasets.

    A GeoDataset is a formalism for gridded data arrays, which are usually
    stored in geotiffs or netcdf files. It provides an interface to realise
    subsets, compute regions of interest and more.

    A GeoDataset makes more sense if it is subclassed for real files,
    such as GeoTiff or GeoNetCDF. In that case, the implemetations must make
    use of the subset indexes provided in the sub_x, sub_y and sub_t
    properties.
    """

    def __init__(self, grid, time=None):
        """Set-up the georeferencing, time is optional.
        Parameters:
        grid: a salem.Grid object which represents the underlying data
        time: if the data has a time dimension
        """

        # The original grid, for always stored
        self._ogrid = grid
        # The current grid (changes if set_subset() is called)
        self.grid = grid
        # Default indexes to get in the underlying data (BOTH inclusive,
        # i.e [, ], not [,[ as in numpy)
        self.sub_x = [0, grid.nx-1]
        self.sub_y = [0, grid.ny-1]
        # Roi is a ny, nx array if set
        self.roi = None
        self.set_roi()

        # _time is a pd.Series because it's so nice to misuse the series.loc
        # flexibility (see set_period)
        if time is not None:
            if isinstance(time, pd.Series):
                time = pd.Series(np.arange(len(time)), index=time.index)
            else:
                try:
                    time = pd.Series(np.arange(len(time)), index=time)
                except AttributeError:
                    # https://github.com/pandas-dev/pandas/issues/23419
                    for t in time:
                        setattr(t, 'nanosecond', 0)
                    time = pd.Series(np.arange(len(time)), index=time)
        self._time = time

        # set_period() will set those
        self.t0 = None
        self.t1 = None
        self.sub_t = None
        self.set_period()

    @property
    def time(self):
        """Time array"""
        if self._time is None:
            return None
        return self._time[self.t0:self.t1].index

    def set_period(self, t0=[0], t1=[-1]):
        """Set a period of interest for the dataset.
         This will be remembered at later calls to time() or GeoDataset's
         getvardata implementations.
         Parameters
         ----------
         t0: anything that represents a time. Could be a string (e.g
         '2012-01-01'), a DateTime, or an index in the dataset's time
         t1: same as t0 (inclusive)
         """

        if self._time is not None:
            # we dont check for what t0 or t1 is, we let Pandas do the job
            # TODO quick and dirty solution for test_longtime, TBR
            self.sub_t = [_to_scalar(self._time[t0]),
                          _to_scalar(self._time[t1])]
            self.t0 = self._time.index[self.sub_t[0]]
            self.t1 = self._time.index[self.sub_t[1]]

    def set_subset(self, corners=None, crs=wgs84, toroi=False, margin=0):
        """Set a subset for the dataset.
         This will be remembered at later calls to GeoDataset's
         getvardata implementations.
         Parameters
         ----------
         corners: a ((x0, y0), (x1, y1)) tuple of the corners of the square
         to subset the dataset to. The coordinates are not expressed in
         wgs84, set the crs keyword
         crs: the coordinates of the corner coordinates
         toroi: set to true to generate the smallest possible subset arond
         the region of interest set with set_roi()
         margin: when doing the subset, add a margin (can be negative!). Can
         be used alone: set_subset(margin=-5) will remove five pixels from
         each boundary of the dataset.
         TODO: shouldnt we make the toroi stuff easier to use?
         """

        # Useful variables
        mx = self._ogrid.nx-1
        my = self._ogrid.ny-1
        cgrid = self._ogrid.center_grid

        # Three possible cases
        if toroi:
            if self.roi is None or np.max(self.roi) == 0:
                raise RuntimeError('roi is empty.')
            ids = np.nonzero(self.roi)
            sub_x = [np.min(ids[1])-margin, np.max(ids[1])+margin]
            sub_y = [np.min(ids[0])-margin, np.max(ids[0])+margin]
        elif corners is not None:
            xy0, xy1 = corners
            x0, y0 = cgrid.transform(*xy0, crs=crs, nearest=True)
            x1, y1 = cgrid.transform(*xy1, crs=crs, nearest=True)
            sub_x = [np.min([x0, x1])-margin, np.max([x0, x1])+margin]
            sub_y = [np.min([y0, y1])-margin, np.max([y0, y1])+margin]
        else:
            # Reset
            sub_x = [0-margin, mx+margin]
            sub_y = [0-margin, my+margin]

        # Some necessary checks
        if (np.max(sub_x) < 0) or (np.min(sub_x) > mx) or \
           (np.max(sub_y) < 0) or (np.min(sub_y) > my):
            raise RuntimeError('subset not valid')

        if (sub_x[0] < 0) or (sub_x[1] > mx):
            warnings.warn('x0 out of bounds', RuntimeWarning)
        if (sub_y[0] < 0) or (sub_y[1] > my):
            warnings.warn('y0 out of bounds', RuntimeWarning)

        # Make the new grid
        sub_x = np.clip(sub_x, 0, mx)
        sub_y = np.clip(sub_y, 0, my)
        nxny = (sub_x[1] - sub_x[0] + 1, sub_y[1] - sub_y[0] + 1)
        dxdy = (self._ogrid.dx, self._ogrid.dy)
        xy0 = (self._ogrid.x0 + sub_x[0] * self._ogrid.dx,
               self._ogrid.y0 + sub_y[0] * self._ogrid.dy)
        self.grid = Grid(proj=self._ogrid.proj, nxny=nxny, dxdy=dxdy, x0y0=xy0)
        # If we arrived here, we can safely set the subset
        self.sub_x = sub_x
        self.sub_y = sub_y

    def set_roi(self, shape=None, geometry=None, crs=wgs84, grid=None,
                corners=None, noerase=False):
        """Set a region of interest for the dataset.
        If set succesfully, a ROI is simply a mask of the same size as the
        dataset's grid, obtained with the .roi attribute.
        I haven't decided yet if the data should be masekd out when a ROI
        has been set.
        Parameters
        ----------
        shape: path to a shapefile
        geometry: a shapely geometry
        crs: the crs of the geometry
        grid: a Grid object
        corners: a ((x0, y0), (x1, y1)) tuple of the corners of the square
        to subset the dataset to. The coordinates are not expressed in
        wgs84, set the crs keyword
        noerase: set to true in order to add the new ROI to the previous one
        """

        # The rois are always defined on the original grids, but of course
        # we take that into account when a subset is set (see roi
        # decorator below)
        ogrid = self._ogrid

        # Initial mask
        if noerase and (self.roi is not None):
            mask = self.roi
        else:
            mask = np.zeros((ogrid.ny, ogrid.nx), dtype=np.int16)

        # Several cases
        if shape is not None:
            if isinstance(shape, pd.DataFrame):
                gdf = shape
            else:
                gdf = sio.read_shapefile(shape)
            gis.transform_geopandas(gdf, to_crs=ogrid.corner_grid,
                                    inplace=True)
            if rasterio is None:
                raise ImportError('This feature needs rasterio')
            from rasterio.features import rasterize
            with rasterio.Env():
                mask = rasterize(gdf.geometry, out=mask)
        if geometry is not None:
            geom = gis.transform_geometry(geometry, crs=crs,
                                          to_crs=ogrid.corner_grid)
            if rasterio is None:
                raise ImportError('This feature needs rasterio')
            from rasterio.features import rasterize
            with rasterio.Env():
                mask = rasterize(np.atleast_1d(geom), out=mask)
        if grid is not None:
            _tmp = np.ones((grid.ny, grid.nx), dtype=np.int16)
            mask = ogrid.map_gridded_data(_tmp, grid, out=mask).filled(0)
        if corners is not None:
            cgrid = self._ogrid.center_grid
            xy0, xy1 = corners
            x0, y0 = cgrid.transform(*xy0, crs=crs, nearest=True)
            x1, y1 = cgrid.transform(*xy1, crs=crs, nearest=True)
            mask[np.min([y0, y1]):np.max([y0, y1])+1,
                 np.min([x0, x1]):np.max([x0, x1])+1] = 1

        self.roi = mask

    @property
    def roi(self):
        """Mask of the ROI (same size as subset)."""
        return self._roi[self.sub_y[0]:self.sub_y[1]+1,
                         self.sub_x[0]:self.sub_x[1]+1]

    @roi.setter
    def roi(self, value):
        """A mask is allways defined on _ogrid"""
        self._roi = value

    def get_vardata(self, var_id=None):
        """Interface to implement by subclasses, taking sub_x, sub_y and
        sub_t into account."""
        raise NotImplementedError()


class GeoTiff(GeoDataset):
    """Geolocalised tiff images (needs rasterio)."""

    def __init__(self, file):
        """Open the file.

        Parameters
        ----------
        file: path to the file
        """
        if rasterio is None:
            raise ImportError('This feature needs rasterio to be insalled')

        # brutally efficient
        with rasterio.Env():
            with rasterio.open(file) as src:
                nxny = (src.width, src.height)
                ul_corner = (src.bounds.left, src.bounds.top)
                proj = pyproj.Proj(src.crs)
                dxdy = (src.res[0], -src.res[1])
                grid = Grid(x0y0=ul_corner, nxny=nxny, dxdy=dxdy,
                            pixel_ref='corner', proj=proj)
        # done
        self.file = file
        GeoDataset.__init__(self, grid)

    def get_vardata(self, var_id=1):
        """Read the geotiff band.

        Parameters
        ----------
        var_id: the variable name (here the band number)
        """
        wx = (self.sub_x[0], self.sub_x[1]+1)
        wy = (self.sub_y[0], self.sub_y[1]+1)
        with rasterio.Env():
            with rasterio.open(self.file) as src:
                band = src.read(var_id, window=(wy, wx))
        return band


class EsriITMIX(GeoDataset):
    """Open ITMIX geolocalised Esri ASCII images (needs rasterio)."""

    def __init__(self, file):
        """Open the file.

        Parameters
        ----------
        file: path to the file
        """

        bname = os.path.basename(file).split('.')[0]
        pok = bname.find('UTM')
        if pok == -1:
            raise ValueError(file + ' does not seem to be an ITMIX file.')
        zone = int(bname[pok+3:])
        south = False
        if zone < 0:
            south = True
            zone = -zone
        proj = pyproj.Proj(proj='utm', zone=zone, ellps='WGS84',
                           south=south)

        # brutally efficient
        with rasterio.Env():
            with rasterio.open(file) as src:
                nxny = (src.width, src.height)
                ul_corner = (src.bounds.left, src.bounds.top)
                dxdy = (src.res[0], -src.res[1])
                grid = Grid(x0y0=ul_corner, nxny=nxny, dxdy=dxdy,
                            pixel_ref='corner', proj=proj)
        # done
        self.file = file
        GeoDataset.__init__(self, grid)

    def get_vardata(self, var_id=1):
        """Read the geotiff band.

        Parameters
        ----------
        var_id: the variable name (here the band number)
        """
        wx = (self.sub_x[0], self.sub_x[1]+1)
        wy = (self.sub_y[0], self.sub_y[1]+1)
        with rasterio.Env():
            with rasterio.open(self.file) as src:
                band = src.read(var_id, window=(wy, wx))
        return band


class GeoNetcdf(GeoDataset):
    """NetCDF files with geolocalisation info.

    GeoNetcdf will try hard to understand the geoloc and time of the file,
    but if it can't you can still provide the time and grid at instantiation.
    """

    def __init__(self, file, grid=None, time=None, monthbegin=False):
        """Open the file and try to understand it.

        Parameters
        ----------
        file: path to the netcdf file
        grid: a Grid object. This will override the normal behavior of
        GeoNetcdf, which is to try to understand the grid automatically.
        time: a time array. This will override the normal behavior of
        GeoNetcdf, which is to try to understand the time automatically.
        monthbegin: set to true if you are sure that your data is monthly
        and that the data provider decided to tag the date as the center of
        the month (stupid)
        """

        self._nc = netCDF4.Dataset(file)
        self._nc.set_auto_mask(False)
        self.variables = self._nc.variables
        if grid is None:
            grid = sio.grid_from_dataset(self._nc)
            if grid is None:
                raise RuntimeError('File grid not understood')
        if time is None:
            time = sio.netcdf_time(self._nc, monthbegin=monthbegin)
        dn = self._nc.dimensions.keys()
        try:
            self.x_dim = utils.str_in_list(dn, utils.valid_names['x_dim'])[0]
            self.y_dim = utils.str_in_list(dn, utils.valid_names['y_dim'])[0]
        except IndexError:
            raise RuntimeError('File coordinates not understood')
        dim = utils.str_in_list(dn, utils.valid_names['t_dim'])
        self.t_dim = dim[0] if dim else None
        dim = utils.str_in_list(dn, utils.valid_names['z_dim'])
        self.z_dim = dim[0] if dim else None

        GeoDataset.__init__(self, grid, time=time)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self.close()

    def close(self):
        self._nc.close()

    def get_vardata(self, var_id=0, as_xarray=False):
        """Reads the data out of the netCDF file while taking into account
        time and spatial subsets.

        Parameters
        ----------
        var_id: the name of the variable (must be available in self.variables)
        as_xarray: returns a DataArray object
        """

        v = self.variables[var_id]

        # Make the slices
        item = []
        for d in v.dimensions:
            it = slice(None)
            if d == self.t_dim and self.sub_t is not None:
                it = slice(self.sub_t[0], self.sub_t[1]+1)
            elif d == self.y_dim:
                it = slice(self.sub_y[0], self.sub_y[1]+1)
            elif d == self.x_dim:
                it = slice(self.sub_x[0], self.sub_x[1]+1)
            item.append(it)

        with np.errstate(invalid='ignore'):
            # This is due to some numpy warnings
            out = v[tuple(item)]

        if as_xarray:
            # convert to xarray
            dims = v.dimensions
            coords = dict()
            x, y = self.grid.x_coord, self.grid.y_coord
            for d in dims:
                if d == self.t_dim:
                    coords[d] = self.time
                elif d == self.y_dim:
                    coords[d] = y
                elif d == self.x_dim:
                    coords[d] = x
            attrs = v.__dict__.copy()
            bad_keys = ['scale_factor', 'add_offset',
                        '_FillValue', 'missing_value', 'ncvars']
            _ = [attrs.pop(b, None) for b in bad_keys]
            out = xr.DataArray(out, dims=dims, coords=coords, attrs=attrs)

        return out


class WRF(GeoNetcdf):
    """WRF proof-of-concept template.

    Adds unstaggered and diagnostic variables.
    """

    def __init__(self, file, grid=None, time=None):

        GeoNetcdf.__init__(self, file, grid=grid, time=time)

        # Change staggered variables to unstaggered ones
        for vn, v in self.variables.items():
            if wrftools.Unstaggerer.can_do(v):
                self.variables[vn] = wrftools.Unstaggerer(v)

        # Check if we can add diagnostic variables to the pot
        for vn in wrftools.var_classes:
            cl = getattr(wrftools, vn)
            if cl.can_do(self._nc):
                self.variables[vn] = cl(self._nc)


class GoogleCenterMap(GeoDataset):
    """Google static map centered on a point.

    Needs motionless.
    """

    def __init__(self, center_ll=(11.38, 47.26), size_x=640, size_y=640,
                 scale=1, zoom=12, maptype='satellite', use_cache=True, 
                 **kwargs):
        """Initialize

        Parameters
        ----------
        center_ll : tuple
          tuple of lon, lat center of the map
        size_x : int
          image size
        size_y : int
          image size
        scale : int
          image scaling factor. 1, 2. 2 is higher resolution but takes 
          longer to download
        zoom : int
          google zoom level (https://developers.google.com/maps/documentation/
          static-maps/intro#Zoomlevels). 1 (world) - 20 (buildings)
        maptype : str, default: 'satellite'
          'roadmap', 'satellite', 'hybrid', 'terrain'
        use_cache : bool, default: True
          store the downloaded image in the cache to avoid future downloads
        kwargs : **
          any keyword accepted by motionless.CenterMap (e.g. `key` for the API)
        """

        global API_KEY

        # Google grid
        grid = gis.googlestatic_mercator_grid(center_ll=center_ll,
                                              nx=size_x, ny=size_y,
                                              zoom=zoom, scale=scale)

        if 'key' not in kwargs:
            if API_KEY is None:
                with open(utils.get_demo_file('.api_key'), 'r') as f:
                    API_KEY = f.read().replace('\n', '')
            kwargs['key'] = API_KEY

        # Motionless
        import motionless
        googleurl = motionless.CenterMap(lon=center_ll[0], lat=center_ll[1],
                                         size_x=size_x, size_y=size_y,
                                         maptype=maptype, zoom=zoom, scale=scale, 
                                         **kwargs)

        # done
        self.googleurl = googleurl
        self.use_cache = use_cache
        GeoDataset.__init__(self, grid)

    @lazy_property
    def _img(self):
        """Download the image."""
        if self.use_cache:
            return utils.joblib_read_img_url(self.googleurl.generate_url())
        else:
            from matplotlib.image import imread
            fd = urlopen(self.googleurl.generate_url())
            return imread(io.BytesIO(fd.read()))

    def get_vardata(self, var_id=0):
        """Return and subset the image."""
        return self._img[self.sub_y[0]:self.sub_y[1]+1,
                         self.sub_x[0]:self.sub_x[1]+1, :]


class GoogleVisibleMap(GoogleCenterMap):
    """Google static map automatically sized and zoomed to a selected region.

    It's usually more practical to use than GoogleCenterMap.
    """

    def __init__(self, x, y, crs=wgs84, size_x=640, size_y=640, scale=1,
                 maptype='satellite', use_cache=True, **kwargs):
        """Initialize

        Parameters
        ----------
        x : array
          x coordinates of the points to include on the map
        y : array
          y coordinates of the points to include on the map
        crs : proj or Grid
          coordinate reference system of x, y
        size_x : int
          image size
        size_y : int
          image size
        scale : int
          image scaling factor. 1, 2. 2 is higher resolution but takes
          longer to download
        maptype : str, default: 'satellite'
          'roadmap', 'satellite', 'hybrid', 'terrain'
        use_cache : bool, default: True
          store the downloaded image in the cache to avoid future downloads
        kwargs : **
          any keyword accepted by motionless.CenterMap (e.g. `key` for the API)

        Notes
        -----
        To obtain the exact domain specified in `x` and `y` you may have to
        play with the `size_x` and `size_y` kwargs.
        """

        global API_KEY

        if 'zoom' in kwargs or 'center_ll' in kwargs:
            raise ValueError('incompatible kwargs.')

        # Transform to lonlat
        crs = gis.check_crs(crs)
        if isinstance(crs, pyproj.Proj):
            lon, lat = gis.transform_proj(crs, wgs84, x, y)
        elif isinstance(crs, Grid):
            lon, lat = crs.ij_to_crs(x, y, crs=wgs84)
        else:
            raise NotImplementedError()

        # surely not the smartest way to do but should be enough for now
        mc = (np.mean(lon), np.mean(lat))
        zoom = 20
        while zoom >= 0:
            grid = gis.googlestatic_mercator_grid(center_ll=mc, nx=size_x,
                                                  ny=size_y, zoom=zoom,
                                                  scale=scale)
            dx, dy = grid.transform(lon, lat, maskout=True)
            if np.any(dx.mask):
                zoom -= 1
            else:
                break

        if 'key' not in kwargs:
            if API_KEY is None:
                with open(utils.get_demo_file('.api_key'), 'r') as f:
                    API_KEY = f.read().replace('\n', '')
            kwargs['key'] = API_KEY

        GoogleCenterMap.__init__(self, center_ll=mc, size_x=size_x,
                                 size_y=size_y, zoom=zoom, scale=scale,
                                 maptype=maptype, use_cache=use_cache, **kwargs)



"""
Color handling and maps.
"""
from __future__ import division
import six

# Builtins
import warnings
import os
from os import path
import copy
# External libs
import numpy as np

try:
    from skimage.transform import resize as imresize
    has_skimage = True
except ImportError:
    has_skimage = False

import pandas as pd

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from mpl_toolkits.axes_grid1 import make_axes_locatable
from matplotlib.collections import PatchCollection, LineCollection
from shapely.geometry import MultiPoint, LineString, Polygon
from descartes.patch import PolygonPatch
from matplotlib.transforms import Transform as MPLTranform

from salem import utils, gis, sio, Grid, wgs84, sample_data_dir, GeoTiff

shapefiles = dict()
shapefiles['world_borders'] = path.join(sample_data_dir, 'shapes',
                                        'world_borders', 'world_borders.shp')
shapefiles['oceans'] = path.join(sample_data_dir, 'shapes', 'oceans',
                                 'ne_50m_ocean.shp')
shapefiles['rivers'] = path.join(sample_data_dir, 'shapes', 'rivers',
                                 'ne_50m_rivers_lake_centerlines.shp')
shapefiles['lakes'] = path.join(sample_data_dir, 'shapes', 'lakes',
                                'ne_50m_lakes.shp')

# Be sure we have the directory
if ~ os.path.exists(shapefiles['world_borders']):
    from salem.utils import get_demo_file
    _ = get_demo_file('world_borders.shp')


class ExtendedNorm(mpl.colors.BoundaryNorm):
    """ A better BoundaryNorm with an ``extend'' keyword.

    TODO: remove this when PR is accepted

    See: https://github.com/matplotlib/matplotlib/issues/4850
         https://github.com/matplotlib/matplotlib/pull/5034
    """

    def __init__(self, boundaries, ncolors, extend='neither'):

        _b = np.atleast_1d(boundaries).astype(float)
        mpl.colors.BoundaryNorm.__init__(self, _b, ncolors, clip=False)

        # 'neither' | 'both' | 'min' | 'max'
        if extend == 'both':
            _b = np.append(_b, _b[-1]+1)
            _b = np.insert(_b, 0, _b[0]-1)
        elif extend == 'min':
            _b = np.insert(_b, 0, _b[0]-1)
        elif extend == 'max':
            _b = np.append(_b, _b[-1]+1)
        self._b = _b
        self._N = len(self._b)
        if self._N - 1 == self.Ncmap:
            self._interp = False
        else:
            self._interp = True

    def __call__(self, value):
        xx, is_scalar = self.process_value(value)
        mask = np.ma.getmaskarray(xx)
        xx = np.atleast_1d(xx.filled(self.vmax + 1))
        iret = np.zeros(xx.shape, dtype=np.int16)
        for i, b in enumerate(self._b):
            iret[xx >= b] = i
        if self._interp:
            scalefac = float(self.Ncmap - 1) / (self._N - 2)
            iret = (iret * scalefac).astype(np.int16)
        iret[xx < self.vmin] = -1
        iret[xx >= self.vmax] = self.Ncmap
        ret = np.ma.array(iret, mask=mask)
        if is_scalar:
            ret = int(ret[0])  # assume python scalar
        return ret


def get_cmap(cmap='viridis'):
    """Get a colormap from mpl, and also those defined by Salem.

    Currently we have: topo, dem, nrwc

    see https://github.com/fmaussion/salem-sample-data/tree/master/colormaps
    """

    try:
        return plt.get_cmap(cmap)
    except ValueError:
        cl = utils.read_colormap(cmap)
        return LinearSegmentedColormap.from_list(cmap, cl, N=256)


class DataLevels(object):
    """Assigns the right color to your data.

    Simple tool that ensures the full compatibility of the plot
    colors and the colorbar. It's working principle is best understood
    with a few examples, available in the docs.
    """

    def __init__(self, data=None, levels=None, nlevels=None, vmin=None,
                 vmax=None, extend=None, cmap=None):
        """Instanciate.

        Parameters
        ----------
        see the set_* functions
        """
        self.set_data(data)
        self.set_levels(levels)
        self.set_nlevels(nlevels)
        self.set_vmin(vmin)
        self.set_vmax(vmax)
        self.set_extend(extend)
        self.set_cmap(cmap)

    def update(self, d):
        """
        Update the properties of :class:`DataLevels` from the dictionary *d*.
        """

        for k, v in six.iteritems(d):
            func = getattr(self, 'set_' + k, None)
            if func is None or not six.callable(func):
                raise AttributeError('Unknown property %s' % k)
            func(v)

    def set_data(self, data=None):
        """Any kind of data array (also masked)."""
        if data is not None:
            self.data = np.ma.masked_invalid(np.atleast_1d(data), copy=False)
        else:
            self.data = np.ma.asarray([0., 1.])

    def set_levels(self, levels=None):
        """Levels you define. Must be monotically increasing."""
        self._levels = levels

    def set_nlevels(self, nlevels=None):
        """Automatic N levels. Ignored if set_levels has been set."""
        self._nlevels = nlevels

    def set_vmin(self, val=None):
        """Mininum level value. Ignored if set_levels has been set."""
        self._vmin = val

    def set_vmax(self, val=None):
        """Maximum level value. Ignored if set_levels has been set."""
        self._vmax = val

    def set_cmap(self, cm=None):
        """Set a colormap."""
        self.cmap = get_cmap(cm or 'viridis' )

    def set_extend(self, extend=None):
        """Colorbar extensions: 'neither' | 'both' | 'min' | 'max'"""
        self._extend = extend

    def set_plot_params(self, levels=None, nlevels=None, vmin=None, vmax=None,
                        extend=None, cmap=None):
        """Shortcut to all parameters related to the plot.

        As a side effect, running set_plot_params() without arguments will
        reset the default behavior
        """
        self.set_vmin(vmin)
        self.set_vmax(vmax)
        self.set_levels(levels)
        self.set_nlevels(nlevels)
        self.set_extend(extend)
        if cmap is not None:
            self.set_cmap(cmap)

    @property
    def levels(self):
        """Clever getter."""
        levels = self._levels
        nlevels = self._nlevels
        if levels is not None:
            self.set_vmin(levels[0])
            self.set_vmax(levels[-1])
            return levels
        else:
            if nlevels is None:
                if self.extend in ['max', 'min']:
                    nlevels = self.cmap.N - 1
                elif self.extend in ['both']:
                    nlevels = self.cmap.N - 2
                else:
                    nlevels = self.cmap.N
            if self.vmax == self.vmin:
                return np.linspace(self.vmin, self.vmax+1, nlevels)
            return np.linspace(self.vmin, self.vmax, nlevels)

    @property
    def nlevels(self):
        """Clever getter."""
        return len(self.levels)

    @property
    def vmin(self):
        """Clever getter."""
        if self._vmin is None:
            return np.min(self.data)
        else:
            return self._vmin

    @property
    def vmax(self):
        """Clever getter."""
        if self._vmax is None:
            return np.max(self.data)
        else:
            return self._vmax

    @property
    def extend(self):
        """Clever getter."""
        if self._extend is None:
            # If the user didnt set it, we decide
            maxd, mind = np.max(self.data), np.min(self.data)
            if maxd > self.vmax and mind < self.vmin:
                out = 'both'
            elif maxd > self.vmax:
                out = 'max'
            elif mind < self.vmin:
                out = 'min'
            else:
                out = 'neither'
            return out
        else:
            return self._extend

    @property
    def norm(self):
        """Clever getter."""
        l = self.levels
        e = self.extend
        # Warnings
        if e not in ['both', 'min'] and (np.min(l) > np.min(self.data)):
            warnings.warn('Minimum data out of bounds.', RuntimeWarning)
        if e not in ['both', 'max'] and (np.max(l) < np.max(self.data)):
            warnings.warn('Maximum data out of bounds.', RuntimeWarning)
        try:
            # Added in mpl 3.3.0
            return mpl.colors.BoundaryNorm(l, self.cmap.N, extend=e)
        except TypeError:
            return ExtendedNorm(l, self.cmap.N, extend=e)

    def to_rgb(self):
        """Transform the data to RGB triples."""

        if np.all(self.data.mask):
            # unfortunately the functions  below can't handle this one
            return np.zeros(self.data.shape + (4, ))

        return self.cmap(self.norm(self.data))

    def get_colorbarbase_kwargs(self):
        """If you need to make a colorbar based on a given DataLevel state."""

        # This is a discutable choice: with more than 60 colors (could be
        # less), we assume a continuous colorbar.
        if self.nlevels < 60:
            norm = self.norm
        else:
            norm = mpl.colors.Normalize(vmin=self.vmin, vmax=self.vmax)
        return dict(extend=self.extend, cmap=self.cmap, norm=norm)

    def colorbarbase(self, cax, **kwargs):
        """Returns a ColorbarBase to add to the cax axis. All keywords are
        passed to matplotlib.colorbar.ColorbarBase
        """

        # This is a discutable choice: with more than 60 colors (could be
        # less), we assume a continuous colorbar.
        if self.nlevels < 60:
            norm = self.norm
        else:
            norm = mpl.colors.Normalize(vmin=self.vmin, vmax=self.vmax)
        return mpl.colorbar.ColorbarBase(cax, extend=self.extend,
                                         cmap=self.cmap, norm=norm, **kwargs)

    def append_colorbar(self, ax, position='right', size='5%', pad=0.5,
                        **kwargs):
        """Appends a colorbar to existing axes

        It uses matplotlib's make_axes_locatable toolkit.

        Parameters
        ----------
        ax: the axis to append the colorbar to
        position: "left"|"right"|"bottom"|"top"
        size: the size of the colorbar (e.g. in % of the ax)
        pad: pad between axes given in inches or tuple-like of floats,
             (horizontal padding, vertical padding)
        """

        orientation = 'horizontal'
        if position in ['left', 'right']:
            orientation = 'vertical'
        cax = make_axes_locatable(ax).append_axes(position, size=size, pad=pad)
        return self.colorbarbase(cax, orientation=orientation, **kwargs)

    def plot(self, ax):
        """Add a kind of plot of the data to an axis.

        More useful for child classes.

        Returns an imshow primitive
        """
        data = np.atleast_2d(self.data)
        toplot = self.cmap(self.norm(data))
        primitive = ax.imshow(toplot, interpolation='none', origin='lower')
        return primitive

    def visualize(self, ax=None, title=None, orientation='vertical',
                  add_values=False, addcbar=True, cbar_title=''):
        """Quick plot, useful for debugging.

        Parameters
        ----------
        ax: the axis to add the plot to (optinal)
        title: the plot title
        orientation: the colorbar's orientation
        add_values: add the data values as text in the pixels (for testing)

        Returns a dict containing the primitives of the various plot calls
        """

        # Do we make our own fig?
        if ax is None:
            ax = plt.gca()

        # Plot
        out = self.plot(ax)

        # Colorbar
        addcbar = (self.vmin != self.vmax) and addcbar
        if addcbar:
            if orientation == 'horizontal':
                self.append_colorbar(ax, "top", size=0.2, pad=0.5,
                                     label=cbar_title)
            else:
                self.append_colorbar(ax, "right", size="5%", pad=0.2,
                                     label=cbar_title)

        # Mini add-on
        if add_values:
            data = np.atleast_2d(self.data)
            x, y = np.meshgrid(np.arange(data.shape[1]),
                               np.arange(data.shape[0]))
            for v, i, j in zip(data.flatten(), x.flatten(), y.flatten()):
                ax.text(i, j, v, horizontalalignment='center',
                        verticalalignment='center')

        # Details
        if title is not None:
            ax.set_title(title)

        return out


class Map(DataLevels):
    """Plotting georeferenced data.

    A Map is an implementation of DataLevels that wraps imshow(), by adding
    all kinds of geoinformation on the plot. It's primary purpose is to add
    country borders or topographical shading. Another purpose is
    to be able to plot all kind of georeferenced data on an existing map
    while being sure that the plot is accurate.

    In short: the Map is a higher-level, less wordy and less flexible
    version of cartopy or basemap. It's usefulness is best shown by the
    examples in the doc.

    For worldwide maps or very flexible plots you'd better use cartopy,
    because Salem's Map is constrained by it's primary objective: plotting
    regional maps.
    """

    def __init__(self, grid, nx=500, ny=None, factor=None,
                 countries=True, **kwargs):
        """Make a new map.

        Parameters
        ----------
        grid : salem.Grid
          the grid defining the map
        nx : int
          x resolution (in pixels) of the map
        ny : int
          y resolution (in pixels) of the map (ignored if nx is set)
        factor : float
          shortcut to keep the same image resolution as the grid.
          1 prevents image resizing
        countries : bool
          automatically add country borders to the map (you can do
          it later with a call to set_shapefile)
        kwargs: **
          all keywords accepted by DataLevels
        """

        if factor is not None:
            nx = None
            ny = None

        self.grid = grid.center_grid.regrid(nx=nx, ny=ny, factor=factor)
        self.origin = 'lower' if self.grid.origin == 'lower-left' else 'upper'

        DataLevels.__init__(self, **kwargs)

        self._collections = []
        self._geometries = []
        self._text = []
        self.set_shapefile(countries=countries)
        self.set_lonlat_contours()
        self._shading_base()
        self._rgb = None
        self._contourf_data = None
        self._contour_data = None

    def _check_data(self, data=None, crs=None, interp='nearest',
                    overplot=False):
        """Interpolates the data to the map grid."""

        if crs is None:
            # try xarray
            # TODO: note that this might slow down the plotting a bit
            # if the data already matches the grid...
            try:
                crs = data.salem.grid
            except:
                pass

        data = np.ma.fix_invalid(np.squeeze(data))
        shp = data.shape
        if len(shp) != 2:
            raise ValueError('Data should be 2D.')

        if crs is None:
            # Reform case, but with a sanity check
            if not np.isclose(shp[0] / shp[1], self.grid.ny / self.grid.nx,
                              atol=1e-2):
                raise ValueError('Dimensions of data do not match the map.')

            # need to resize if not same
            if not ((shp[0] == self.grid.ny) and (shp[1] == self.grid.nx)):

                # We convert to float for img resizing
                if data.dtype not in [np.float32, np.float64]:
                    data = data.astype(np.float64)

                if interp.lower() == 'nearest':
                    interp = 0
                elif interp.lower() == 'linear':
                    interp = 1
                elif interp.lower() == 'spline':
                    interp = 3
                if not has_skimage:
                    raise ImportError('Needs scikit-image to be installed.')
                with warnings.catch_warnings():
                    mess = "invalid value encountered in reduce"
                    warnings.filterwarnings("ignore", message=mess)
                    mess = ("Possible precision loss when converting from "
                            "int64 to float64")
                    warnings.filterwarnings("ignore", message=mess)
                    try:
                        data = imresize(data.filled(np.NaN),
                                        (self.grid.ny, self.grid.nx),
                                        order=interp, mode='edge',
                                        anti_aliasing=True)
                    except RuntimeError:
                        # For some order anti_aliasing doesnt work with 'edge'
                        data = imresize(data.filled(np.NaN),
                                        (self.grid.ny, self.grid.nx),
                                        order=interp, mode='edge',
                                        anti_aliasing=False)

            return data

        crs = gis.check_crs(crs, raise_on_error=True)
        if isinstance(crs, Grid):
            # Remap
            if overplot:
                data = self.grid.map_gridded_data(data, crs, interp=interp,
                                                  out=self.data)
            else:
                data = self.grid.map_gridded_data(data, crs, interp=interp)
        else:
            raise ValueError('crs should be a grid, not a proj')

        return data

    def set_data(self, data=None, crs=None, interp='nearest',
                 overplot=False):
        """Adds data to the plot. The data has to be georeferenced, i.e. by
        setting crs (if omitted the data is assumed to be defined on the
        map's grid)

        Parameters
        ----------
        data: the data array (2d)
        crs: the data coordinate reference system
        interp: 'nearest' (default) or 'linear', the interpolation algorithm
        overplot: add the data to an existing plot (useful for mosaics for
        example)
        """

        # Check input
        if data is None:
            self.data = np.ma.zeros((self.grid.ny, self.grid.nx))
            self.data.mask = self.data + 1
            return
        data = self._check_data(data=data, crs=crs, interp=interp,
                                overplot=overplot)
        DataLevels.set_data(self, data)

    def set_contourf(self, data=None, crs=None, interp='nearest', **kwargs):
        """Adds data to contourfill on the map.

        Parameters
        ----------
        mask: bool array (2d)
        crs: the data coordinate reference system
        interp: 'nearest' (default) or 'linear', the interpolation algorithm
        kwargs: anything accepted by contourf
        """


        # Check input
        if data is None:
            self._contourf_data = None
            return

        self._contourf_data = self._check_data(data=data, crs=crs,
                                               interp=interp)
        kwargs.setdefault('zorder', 1.4)
        self._contourf_kw = kwargs

    def set_contour(self, data=None, crs=None, interp='nearest', **kwargs):
        """Adds data to contour on the map.

        Parameters
        ----------
        mask: bool array (2d)
        crs: the data coordinate reference system
        interp: 'nearest' (default) or 'linear', the interpolation algorithm
        kwargs: anything accepted by contour
        """


        # Check input
        if data is None:
            self._contour_data = None
            return

        self._contour_data = self._check_data(data=data, crs=crs,
                                              interp=interp)
        kwargs.setdefault('zorder', 1.4)
        self._contour_kw = kwargs

    def set_geometry(self, geometry=None, crs=wgs84, text=None,
                     text_delta=(0.01, 0.01), text_kwargs=dict(), **kwargs):
        """Adds any Shapely geometry to the map.

        If called without arguments, it removes all previous geometries.

        Parameters
        ----------
        geometry: a Shapely gometry object (must be a scalar!)
        crs: the associated coordinate reference system (default wgs84)
        text: if you want to add a text to the geometry (it's position is
        based on the geometry's centroid)
        text_delta: it can be useful to shift the text of a certain amount
        when annotating points. units are percentage of data coordinates.
        text_kwargs: the keyword arguments to pass to the test() function
        kwargs: any keyword associated with the geometry's plotting function:

            - Point: all keywords accepted by scatter(): marker, s, edgecolor,
              facecolor...
            - Line: all keywords accepted by plot(): color, linewidth...
            - Polygon: all keywords accepted by PathPatch(): color, edgecolor,
              facecolor, linestyle, linewidth, alpha...

        """

        # Reset?
        if geometry is None:
            self._geometries = []
            return

        # Transform
        geom = gis.transform_geometry(geometry, crs=crs,
                                      to_crs=self.grid.center_grid)

        # Text
        if text is not None:
            x, y = geom.centroid.xy
            x = x[0] + text_delta[0] * self.grid.nx
            sign = self.grid.dy / np.abs(self.grid.dy)
            y = y[0] + text_delta[1] * self.grid.ny * sign
            self.set_text(x, y, text, crs=self.grid.center_grid,
                          **text_kwargs)

        # Save
        if 'Multi' in geom.type:
            for g in geom:
                self._geometries.append((g, kwargs))
                # dirty solution: I should use collections instead
                if 'label' in kwargs:
                    kwargs = kwargs.copy()
                    del kwargs['label']
        else:
            self._geometries.append((geom, kwargs))

    def set_points(self, x, y, **kwargs):
        """Shortcut for set_geometry() accepting coordinates as input."""
        x, y = np.atleast_1d(x), np.atleast_1d(y)
        self.set_geometry(MultiPoint(np.array([x, y]).T), **kwargs)

    def set_text(self, x=None, y=None, text='', crs=wgs84, **kwargs):
        """Add a text to the map.

        Keyword arguments will be passed to mpl's text() function.
        """

        # Reset?
        if x is None:
            self._text = []
            return

        kwargs.setdefault('zorder', 5)

        # Transform
        x, y = self.grid.center_grid.transform(x, y, crs=crs)
        self._text.append((x, y, text, kwargs))

    def set_shapefile(self, shape=None, countries=False, oceans=False,
                      rivers=False, lakes=False, **kwargs):
        """Add a shapefile to the plot.

        Salem is shipped with a few default settings for country borders,
        oceans and rivers (set one at the time!)

        set_shapefile() without argument will reset the map to zero shapefiles.

        Parameters
        ----------
        shape: the path to the shapefile to read
        countries: if True, add country borders
        oceans: if True, add oceans
        rivers: if True, add rivers
        lakes: if True, add lakes
        kwargs: all keywords accepted by the corresponding collection.
        For LineStrings::
            linewidths, colors, linestyles, ...
        For Polygons::
            alpha, edgecolor, facecolor, fill, linestyle, linewidth, color, ...
        """

        # See if the user wanted defaults settings
        if oceans:
            kwargs.setdefault('facecolor', (0.36862745, 0.64313725, 0.8))
            kwargs.setdefault('edgecolor', 'none')
            kwargs.setdefault('alpha', 1)
            return self.set_shapefile(shapefiles['oceans'], **kwargs)
        if rivers:
            kwargs.setdefault('color', (0.08984375, 0.65625, 0.8515625))
            return self.set_shapefile(shapefiles['rivers'], **kwargs)
        if lakes:
            kwargs.setdefault('color', (0.08984375, 0.65625, 0.8515625))
            return self.set_shapefile(shapefiles['lakes'], **kwargs)
        if countries:
            kwargs.setdefault('zorder', 1.5)
            return self.set_shapefile(shapefiles['world_borders'], **kwargs)

        # Defaults
        if not all(k in kwargs for k in ('facecolor', 'edgecolor')):
            kwargs.setdefault('color', 'k')

        # Reset?
        if shape is None:
            self._collections = []
            return

        # Transform
        if isinstance(shape, pd.DataFrame):
            shape = gis.transform_geopandas(shape, to_crs=self.grid)
        else:
            shape = sio.read_shapefile_to_grid(shape, grid=self.grid)
        if len(shape) == 0:
            return

        # Different collection for each type
        geomtype = shape.iloc[0].geometry.type
        if 'Polygon' in geomtype:
            patches = []
            for g in shape.geometry:
                if 'Multi' in g.type:
                    for gg in g:
                        patches.append(PolygonPatch(gg))
                else:
                    patches.append(PolygonPatch(g))
            kwargs.setdefault('facecolor', 'none')
            if 'color' in kwargs:
                kwargs.setdefault('edgecolor', kwargs['color'])
                del kwargs['color']
            self._collections.append(PatchCollection(patches, **kwargs))
        elif 'LineString' in geomtype:
            lines = []
            for g in shape.geometry:
                if 'Multi' in g.type:
                    for gg in g:
                        lines.append(np.array(gg))
                else:
                    lines.append(np.array(g))
            self._collections.append(LineCollection(lines, **kwargs))
        else:
            raise NotImplementedError(geomtype)

    def _find_interval(self, max_nticks):
        """Quick n dirty function to find a suitable lonlat interval."""
        candidates = [0.001, 0.002, 0.005,
                      0.01, 0.02, 0.05,
                      0.1, 0.2, 0.5,
                      1, 2, 5, 10, 20]
        xx, yy = self.grid.pixcorner_ll_coordinates
        for inter in candidates:
            _xx = xx / inter
            _yy = yy / inter
            mm_x = [np.ceil(np.min(_xx)), np.floor(np.max(_xx))]
            mm_y = [np.ceil(np.min(_yy)), np.floor(np.max(_yy))]
            nx = mm_x[1]-mm_x[0]+1
            ny = mm_y[1]-mm_y[0]+1
            if np.max([nx, ny]) <= max_nticks:
                break
        return inter

    def set_lonlat_contours(self, interval=None, xinterval=None,
                            yinterval=None, add_tick_labels=True,
                            add_xtick_labels=True, add_ytick_labels=True,
                            max_nticks=8, **kwargs):
        """Add longitude and latitude contours to the map.

        Calling it with interval=0 will remove all contours.

        Parameters
        ----------
        interval : float
            interval (in degrees) between the contours (same for lon and lat)
        xinterval : float
            set a different interval for lons
        yinterval : float
            set a different interval for lats
        add_tick_label : bool
            add ticks labels to the axes
        add_xtick_label : bool
            add ticks labels to the x axis
        add_ytick_label : bool
            add ticks labels to the y axis
        max_nticks : int
            maximum number of contour levels (when chosen automatically). 
            Ignore if ``interval`` is set to a value 
        kwargs : {}
            any keyword accepted by contour()
        """

        # Defaults
        if interval is None:
            interval = self._find_interval(max_nticks)
        if xinterval is None:
            xinterval = interval
        if yinterval is None:
            yinterval = interval

        if xinterval == 0:
            # no contour
            self.xtick_levs = []
            self.ytick_levs = []
            add_tick_labels = False
        else:
            # Change XY into interval coordinates, and back after rounding
            xx, yy = self.grid.pixcorner_ll_coordinates
            _xx = xx / xinterval
            _yy = yy / yinterval
            mm_x = [np.ceil(np.min(_xx)), np.floor(np.max(_xx))]
            mm_y = [np.ceil(np.min(_yy)), np.floor(np.max(_yy))]
            self.xtick_levs = (mm_x[0] + np.arange(mm_x[1]-mm_x[0]+1)) * \
                              xinterval
            self.ytick_levs = (mm_y[0] + np.arange(mm_y[1]-mm_y[0]+1)) * \
                              yinterval

        # Decide on float format
        d = np.array(['4', '3', '2', '1', '0'])
        d = d[interval < np.array([0.001, 0.01, 0.1, 1, 10000])][0]

        # The labels (quite ugly)
        self.xtick_pos = []
        self.xtick_val = []
        self.ytick_pos = []
        self.ytick_val = []
        if add_tick_labels:
            if add_xtick_labels:
                _xx = xx[0 if self.origin == 'lower' else -1, :]
                _xi = np.arange(self.grid.nx+1)
                for xl in self.xtick_levs:
                    if (xl > _xx[-1]) or (xl < _xx[0]):
                        continue
                    self.xtick_pos.append(np.interp(xl, _xx, _xi))
                    label = ('{:.' + d + 'f}').format(xl)
                    label += 'W' if (xl < 0) else 'E'
                    if xl == 0:
                        label = '0'
                    self.xtick_val.append(label)
            if add_ytick_labels:
                _yy = np.sort(yy[:, 0])
                _yi = np.arange(self.grid.ny+1)
                if self.origin == 'upper':
                    _yi = _yi[::-1]
                for yl in self.ytick_levs:
                    if (yl > _yy[-1]) or (yl < _yy[0]):
                        continue
                    self.ytick_pos.append(np.interp(yl, _yy, _yi))
                    label = ('{:.' + d + 'f}').format(yl)
                    label += 'S' if (yl < 0) else 'N'
                    if yl == 0:
                        label = 'Eq.'
                    self.ytick_val.append(label)

        # Done
        kwargs.setdefault('colors', 'gray')
        kwargs.setdefault('linestyles', 'dashed')
        kwargs.setdefault('linewidths', 0.8)
        kwargs.setdefault('zorder', 1.5)
        self.ll_contour_kw = kwargs

    def _shading_base(self, slope=None, relief_factor=0.7):
        """Compute the shading factor out of the slope."""

        # reset?
        if slope is None:
            self.slope = None
            return

        # I got this formula from D. Scherer. It works and I dont know why
        p = np.nonzero(slope > 0)
        if len(p[0]) > 0:
            temp = np.clip(slope[p] / (2 * np.std(slope)), -1, 1)
            slope[p] = 0.4 * np.sin(0.5*np.pi*temp)
        self.relief_factor = relief_factor
        self.slope = slope

    def set_topography(self, topo=None, crs=None, relief_factor=0.7, **kwargs):
        """Add topographical shading to the map.

        Parameters
        ----------
        topo: path to a geotiff file containing the topography, OR
              2d data array
        relief_factor: how strong should the shading be?
        kwargs: any keyword accepted by salem.Grid.map_gridded_data (interp,ks)

        Returns
        -------
        the topography if needed (bonus)
        """

        if topo is None:
            self._shading_base()
            return

        kwargs.setdefault('interp', 'spline')

        if isinstance(topo, six.string_types):
            _, ext = os.path.splitext(topo)
            if ext.lower() == '.tif':
                g = GeoTiff(topo)
                # Spare memory
                ex = self.grid.extent_in_crs(crs=wgs84)  # l, r, b, t
                g.set_subset(corners=((ex[0], ex[2]), (ex[1], ex[3])),
                             crs=wgs84, margin=10)
                z = g.get_vardata()
                z[z < -999] = 0
                z = self.grid.map_gridded_data(z, g.grid, **kwargs)
            else:
                raise ValueError('File extension not recognised: {}'
                                 .format(ext))
        else:
            z = self._check_data(topo, crs=crs, **kwargs)

        # Gradient in m m-1
        ddx = self.grid.dx
        ddy = self.grid.dy
        if gis.proj_is_latlong(self.grid.proj):
            # we make a coarse approx of the avg dx on a sphere
            _, lat = self.grid.ll_coordinates
            ddx = np.mean(ddx * 111200 * np.cos(lat * np.pi / 180))
            ddy *= 111200

        dy, dx = np.gradient(z, ddy, ddx)
        self._shading_base(dx - dy, relief_factor=relief_factor)
        return z

    def set_rgb(self, img=None, crs=None, interp='nearest',
                natural_earth=None):
        """Manually force to a rgb img

        Parameters
        ----------
        img : array
            the image to plot
        crs : Grid
            the image reference system
        interp : str, default 'nearest'
            'nearest', 'linear', or 'spline'
        natural_earth : str
           'lr', 'mr' or 'hr' (low res, medium or high res)
           natural earth background img
        """

        if natural_earth is not None:
            from matplotlib.image import imread
            with warnings.catch_warnings():
                # DecompressionBombWarning
                warnings.simplefilter("ignore")
                img = imread(utils.get_natural_earth_file(natural_earth))
            ny, nx = img.shape[0], img.shape[1]
            dx, dy = 360. / nx, 180. / ny
            grid = Grid(nxny=(nx, ny), dxdy=(dx, -dy), x0y0=(-180., 90.),
                        pixel_ref='corner').center_grid
            return self.set_rgb(img, grid, interp='linear')

        if (len(img.shape) != 3) or (img.shape[-1] not in [3, 4]):
            raise ValueError('img should be of shape (y, x, 3) or (y, x, 4)')

        # Unefficient but by far easiest right now
        out = []
        for i in range(img.shape[-1]):
            out.append(self._check_data(img[..., i], crs=crs, interp=interp))
        self._rgb = np.dstack(out)

    def set_scale_bar(self, location=None, length=None, maxlen=0.25,
                      add_bbox=False, bbox_dx=1.2, bbox_dy=1.2,
                      bbox_kwargs=None, **kwargs):
        """Add a legend bar showing the scale to the plot.

        Parameters
        ----------
        location : tuple
            the location of the bar (in the plot's relative coordinates)
        length : float
            the length of the bar in proj units (m or deg). Default is to
            find the nicest number satisfying ``maxlen``
        maxlen : float
            the maximum lenght of the bar (in the plot's relative coordinates)
            when choosing the length automatically
        add_bbox : bool
            add a bounding box to the scale bar (WIP: experimental)
        bbox_dx : bool, default: 1.2
            a multiplicating factor controlling the x size of the bounding box
            (trial and error works best for this one)
        bbox_dy : bool, default: 1.2
            a multiplicating factor controlling the y size of the bounding box
            (trial and error works best for this one)
        bbox_kwargs : dict
            kwarg to pass to set_geometry() for the bounding box (e.g.
            facecolor, alpha, etc...)
        kwargs : dict
            any kwarg accepted by ``set_geometry``. Defaults are put on
            ``color``, ``linewidth``, ``text``, ``text_kwargs``... But you can
            do whatever you want
        """

        x0, x1, y0, y1 = self.grid.extent

        # Find a sensible length for the scale
        if length is None:
            length = utils.nice_scale(x1 - x0, maxlen=maxlen)

        if location is None:
            location = (0.96 - length/2/(x1 - x0), 0.04)

        # scalebar center location in proj coordinates
        sbcx, sbcy = x0 + (x1 - x0) * location[0], y0 + (y1 - y0) * location[1]

        # coordinates for the scalebar
        line = LineString(([sbcx - length/2, sbcy], [sbcx + length/2, sbcy]))
        # Of the bounding box
        bbox = [[sbcx - length / 2 * bbox_dx, sbcy - length / 4 * bbox_dy],
                [sbcx - length / 2 * bbox_dx, sbcy + length / 4 * bbox_dy],
                [sbcx + length / 2 * bbox_dx, sbcy + length / 4 * bbox_dy],
                [sbcx + length / 2 * bbox_dx, sbcy - length / 4 * bbox_dy],
                ]

        # Units
        if gis.proj_is_latlong(self.grid.proj):
            units = 'deg'
        elif length >= 1000.:
            length /= 1000
            units = 'km'
        else:
            units = 'm'
        # Nice number
        if int(length) == length:
            length = int(length)
        # Defaults
        kwargs.setdefault('color', 'k')
        kwargs.setdefault('text', '{} '.format(length) + units)
        kwargs.setdefault('text_delta', (0.0, 0.015))
        kwargs.setdefault('linewidth', 3)
        kwargs.setdefault('zorder', 99)
        tkw = kwargs.get('text_kwargs', {})
        tkw.setdefault('horizontalalignment', 'center')
        tkw.setdefault('zorder', 99)
        tkw.setdefault('color', kwargs['color'])
        kwargs['text_kwargs'] = tkw
        if add_bbox:
            if bbox_kwargs is None:
                bbox_kwargs = {}
            bbox_kwargs.setdefault('facecolor', 'w')
            bbox_kwargs.setdefault('edgecolor', 'k')
            bbox_kwargs.setdefault('zorder', 98)
            poly = Polygon(np.asarray(bbox))
            self.set_geometry(poly, crs=self.grid.proj, **bbox_kwargs)
        self.set_geometry(line, crs=self.grid.proj, **kwargs)

    def transform(self, crs=wgs84, ax=None):
        """Get a matplotlib transform object for a given reference system

        Parameters
        ----------
        crs : coordinate reference system
            a Grid or a Proj, basically. If a grid is given, the grid's proj
            will be used.

        Returns
        -------
        a matplotlib.transforms.Transform instance
        """
        try:
            crs = crs.salem.grid
        except:
            pass
        try:
            crs = crs.proj
        except:
            pass
        return _SalemTransform(target_grid=self.grid,
                               source_crs=crs, ax=ax)

    def to_rgb(self):
        """Transform the data to a RGB image and add topographical shading."""

        if self._rgb is None:
            toplot = DataLevels.to_rgb(self)
        else:
            toplot = self._rgb

        # Shading
        if self.slope is not None:
            # remove alphas?
            try:
                pno = np.where(toplot[:, :, 3] == 0.)
                for i in [0, 1, 2]:
                    toplot[pno[0], pno[1], i] = 1
                toplot[:, :, 3] = 1
            except IndexError:
                pass

            # Actual shading
            level = 1.0 - 0.1 * self.relief_factor
            sens = 1 + 0.7 * self.relief_factor * self.slope
            for i in [0, 1, 2]:
                toplot[:, :, i] = np.clip(level * toplot[:, :, i] * sens, 0, 1)

        # OK!
        return toplot

    def plot(self, ax):
        """Add the map plot to an axis.

        It first plots the image and then adds all the cartographic
        information on top of it.

        Returns a dict containing the primitives of the various plot calls
        """

        out = {'imshow': None,
               'contour': [],
               'contourf': [],
               }

        # Image is the easiest
        out['imshow'] = ax.imshow(self.to_rgb(), interpolation='none',
                                  origin=self.origin)
        ax.autoscale(enable=False)

        # Contour
        if self._contour_data is not None:
            cs = ax.contour(self._contour_data, **self._contour_kw)
            out['contour'].append(cs)

        # Contourfill
        if self._contourf_data is not None:
            cs = ax.contourf(self._contourf_data, **self._contourf_kw)
            out['contourf'].append(cs)

        # Shapefiles
        for col in self._collections:
            ax.add_collection(copy.copy(col))

        # Lon lat contours
        lon, lat = self.grid.pixcorner_ll_coordinates
        if len(self.xtick_levs) > 0:
            ax.contour(lon, levels=self.xtick_levs,
                       extent=(-0.5, self.grid.nx-0.5, -0.5, self.grid.ny),
                       **self.ll_contour_kw)
        if len(self.ytick_levs) > 0:
            ax.contour(lat, levels=self.ytick_levs,
                       extent=(-0.5, self.grid.nx, -0.5, self.grid.ny-0.5),
                       **self.ll_contour_kw)

        # Geometries
        for g, kwargs in self._geometries:
            if g.type == 'Polygon':
                kwargs.setdefault('facecolor', 'none')
                plot_polygon(ax, g, **kwargs)  # was g.buffer(0). Why?
            if g.type in ['LineString', 'LinearRing']:
                a = np.array(g)
                kwargs.setdefault('color', 'k')
                ax.plot(a[:, 0], a[:, 1], **kwargs)
            if g.type == 'Point':
                kwargs.setdefault('zorder', 5)
                kwargs.setdefault('marker', 'o')
                kwargs.setdefault('s', 60)
                kwargs.setdefault('facecolor', 'w')
                kwargs.setdefault('edgecolor', 'k')
                kwargs.setdefault('linewidths', 1)
                if 'markersize' in kwargs:
                    # For those tempted to use the whole kw
                    kwargs['s'] = kwargs['markersize']
                    del kwargs['markersize']
                if 'color' in kwargs:
                    # For those tempted to use the whole kw
                    kwargs['facecolor'] = kwargs['color']
                    kwargs['edgecolor'] = kwargs['color']
                if 'c' in kwargs:
                    # For those tempted to use the whole kw
                    kwargs['facecolor'] = kwargs['c']
                    kwargs['edgecolor'] = kwargs['c']
                    del kwargs['c']
                ax.scatter(g.x, g.y, **kwargs)

        # Texts
        for x, y, s, kwargs in self._text:
            ax.text(x, y, s, **kwargs)

        # Ticks
        if (len(self.xtick_pos) > 0) or (len(self.ytick_pos) > 0):
            ax.xaxis.set_ticks(np.array(self.xtick_pos)-0.5)
            ax.yaxis.set_ticks(np.array(self.ytick_pos)-0.5)
            ax.set_xticklabels(self.xtick_val)
            ax.set_yticklabels(self.ytick_val)
            ax.xaxis.set_zorder(5)
            ax.yaxis.set_zorder(5)
        else:
            ax.xaxis.set_ticks([])
            ax.yaxis.set_ticks([])

        return out


def plot_polygon(ax, poly, edgecolor='black', **kwargs):
    """ Plot a single Polygon geometry """

    a = np.asarray(poly.exterior)
    # without Descartes, we could make a Patch of exterior
    ax.add_patch(PolygonPatch(poly, **kwargs))
    ax.plot(a[:, 0], a[:, 1], color=edgecolor)
    for p in poly.interiors:
        x, y = zip(*p.coords)
        ax.plot(x, y, color=edgecolor)


class _SalemTransform(MPLTranform):
    """
    A transform class for mpl axes using Grids.
    """

    input_dims = 2
    output_dims = 2
    is_separable = False
    has_inverse = False

    def __init__(self, target_grid=None, source_crs=None, ax=None):
        """ Instanciate.

        Parameters
        ----------
        target_grid : salem.Grid
            typically, the map grid
        source_grid
        """
        self.source_crs = source_crs
        self.target_grid = target_grid
        self.ax = ax
        MPLTranform.__init__(self)

    def transform_non_affine(self, xy):
        xx, yy = xy[:, 0:1], xy[:, 1:2]
        xx, yy = self.target_grid.transform(xx, yy, crs=self.source_crs)
        ax = plt.gca() if self.ax is None else self.ax
        return ax.transData.transform(np.concatenate((xx, yy), 1))

    def transform_path_non_affine(self, path):
        raise NotImplementedError('path transforms not working yet')

    def inverted(self):
        raise NotImplementedError('inverted not working yet')



"""
Some useful functions
"""
from __future__ import division

import io
import os
import shutil
import zipfile
from collections import OrderedDict

import numpy as np
from joblib import Memory
from salem import (cache_dir, sample_data_dir, sample_data_gh_commit,
                   download_dir, python_version)
from six.moves.urllib.request import urlretrieve, urlopen


def _hash_cache_dir():
    """Get the path to the right cache directory.

    We need to make sure that cached files correspond to the same
    environment. To this end we make a unique directory hash, depending on the
    version and location of several packages we thought are important
    (because they change often, or because conda versions give different
    results than pip versions).

    Returns
    -------
    path to the dir
    """
    import hashlib

    out = OrderedDict(python_version=python_version)

    try:
        import shapely
        out['shapely_version'] = shapely.__version__
        out['shapely_file'] = shapely.__file__
    except ImportError:
        pass
    try:
        import fiona
        out['fiona_version'] = fiona.__version__
        out['fiona_file'] = fiona.__file__
    except ImportError:
        pass
    try:
        import pandas
        out['pandas_version'] = pandas.__version__
        out['pandas_file'] = pandas.__file__
    except ImportError:
        pass
    try:
        import geopandas
        out['geopandas_version'] = geopandas.__version__
        out['geopandas_file'] = geopandas.__file__
    except ImportError:
        pass
    try:
        import osgeo
        out['osgeo_version'] = osgeo.__version__
        out['osgeo_file'] = osgeo.__file__
    except ImportError:
        pass
    try:
        import pyproj
        out['pyproj_version'] = pyproj.__version__
        out['pyproj_file'] = pyproj.__file__
    except ImportError:
        pass
    try:
        import salem
        out['salem_version'] = salem.__version__
        out['salem_file'] = salem.__file__
    except ImportError:
        pass

    # ok, now make a dummy str that we will hash
    strout = ''
    for k, v in out.items():
        strout += k + v
    strout = 'salem_hash_' + hashlib.md5(strout.encode()).hexdigest()
    dirout = os.path.join(cache_dir, 'cache', strout)
    return dirout


hash_cache_dir = _hash_cache_dir()
try:
    memory = Memory(location=hash_cache_dir + '_joblib', verbose=0)
except TypeError:
    # https://github.com/fmaussion/salem/issues/130
    memory = Memory(cachedir=hash_cache_dir + '_joblib', verbose=0)

# A series of variables and dimension names that Salem will understand
valid_names = dict()
valid_names['x_dim'] = ['west_east', 'lon', 'longitude', 'longitudes', 'lons',
                        'xlong', 'xlong_m', 'dimlon', 'x', 'lon_3', 'long',
                        'phony_dim_0', 'eastings', 'easting', 'nlon', 'nlong',
                        'grid_longitude_t']
valid_names['y_dim'] = ['south_north', 'lat', 'latitude', 'latitudes', 'lats',
                        'xlat', 'xlat_m', 'dimlat', 'y','lat_3', 'phony_dim_1',
                        'northings', 'northing', 'nlat', 'grid_latitude_t']
valid_names['z_dim'] = ['levelist','level', 'pressure', 'press', 'zlevel', 'z',
                        'bottom_top']
valid_names['t_dim'] = ['time', 'times', 'xtime']

valid_names['lon_var'] = ['lon', 'longitude', 'longitudes', 'lons', 'long']
valid_names['lat_var'] = ['lat', 'latitude', 'latitudes', 'lats']
valid_names['time_var'] = ['time', 'times']

sample_data_gh_repo = 'fmaussion/salem-sample-data'
nearth_base = 'http://shadedrelief.com/natural3/ne3_data/'


def str_in_list(l1, l2):
    """Check if one element of l1 is in l2 and if yes, returns the name of
    that element in a list (could be more than one.

    Examples
    --------
    >>> print(str_in_list(['time', 'lon'], ['temp','time','prcp']))
    ['time']
    >>> print(str_in_list(['time', 'lon'], ['temp','time','prcp','lon']))
    ['time', 'lon']
    """
    return [i for i in l1 if i.lower() in l2]


def empty_cache():
    """Empty salem's cache directory."""

    if os.path.exists(cache_dir):
        shutil.rmtree(cache_dir)
    os.makedirs(cache_dir)


def cached_shapefile_path(fpath):
    """Checks if a shapefile is cached and returns the corresponding path.

    This function checks for the last time the file has changed,
    so it should be safe to use.
    """

    p, ext = os.path.splitext(fpath)

    if ext.lower() == '.p':
        # No need to recache pickled files (this is for nested calls)
        return fpath

    if ext.lower() != '.shp':
        raise ValueError('File extension not recognised: {}'.format(ext))

    # Cached directory and file
    cp = os.path.commonprefix([cache_dir, p])
    cp = os.path.join(cache_dir, hash_cache_dir + '_shp',
                      os.path.relpath(p, cp))
    ct = '{:d}'.format(int(round(os.path.getmtime(fpath)*1000.)))
    of = os.path.join(cp, ct + '.p')
    if os.path.exists(cp):
        # We have to check if the file changed
        if os.path.exists(of):
            return of
        else:
            # the file has changed
            shutil.rmtree(cp)

    os.makedirs(cp)
    return of


def _urlretrieve(url, ofile, *args, **kwargs):
    """Wrapper for urlretrieve which overwrites."""

    try:
        return urlretrieve(url, ofile, *args, **kwargs)
    except:
        if os.path.exists(ofile):
            os.remove(ofile)
        # try to make the thing more robust with a second shot
        try:
            return urlretrieve(url, ofile, *args, **kwargs)
        except:
            if os.path.exists(ofile):
                os.remove(ofile)
            raise


def download_demo_files():
    """Checks if the demo data is already on the cache and downloads it.

    Borrowed from OGGM.
    """

    master_zip_url = 'https://github.com/%s/archive/%s.zip' % \
                     (sample_data_gh_repo, sample_data_gh_commit)
    ofile = os.path.join(cache_dir,
                         'salem-sample-data-%s.zip' % sample_data_gh_commit)
    odir = os.path.join(cache_dir)

    # download only if necessary
    if not os.path.exists(ofile):
        print('Downloading salem-sample-data...')
        _urlretrieve(master_zip_url, ofile)

        # Trying to make the download more robust
        try:
            with zipfile.ZipFile(ofile) as zf:
                zf.extractall(odir)
        except zipfile.BadZipfile:
            # try another time
            if os.path.exists(ofile):
                os.remove(ofile)
            _urlretrieve(master_zip_url, ofile)
            with zipfile.ZipFile(ofile) as zf:
                zf.extractall(odir)

    # list of files for output
    out = dict()
    for root, directories, filenames in os.walk(sample_data_dir):
        for filename in filenames:
            out[filename] = os.path.join(root, filename)

    return out


def get_demo_file(fname):
    """Returns the path to the desired demo file."""

    d = download_demo_files()
    if fname in d:
        return d[fname]
    else:
        return None


def get_natural_earth_file(res='lr'):
    """Returns the path to the desired natural earth file.

    http://www.shadedrelief.com/natural3/pages/textures.html

    Parameters
    ----------
    res : str
       'lr' or 'hr' (low res or high res)
    """

    if not os.path.exists(download_dir):
        os.makedirs(download_dir)

    if res == 'lr':
        return get_demo_file('natural_earth_lr.jpg')
    elif res == 'mr':
        urlpath = nearth_base + '8192/textures/2_no_clouds_8k.jpg'
    elif res == 'hr':
        urlpath = nearth_base + '16200/textures/2_no_clouds_16k.jpg'
    ofile = os.path.join(download_dir, 'natural_earth_' + res+ '.jpg')

    # download only if necessary
    if not os.path.exists(ofile):
        print('Downloading Natural Earth ' + res + '...')
        _urlretrieve(urlpath, ofile)

    return ofile


@memory.cache
def read_colormap(name):
    """Reads a colormap from the custom files in Salem."""

    path = get_demo_file(name + '.c3g')

    out = []
    with open(path, 'r') as file:
        for line in file:
            if 'rgb(' not in line:
                continue
            line = line.split('(')[-1].split(')')[0]
            out.append([float(n) for n in line.split(',')])

    return np.asarray(out).astype(float) / 256.


@memory.cache
def joblib_read_img_url(url):
    """Prevent to re-download from GoogleStaticMap if it was done before"""

    from matplotlib.image import imread
    fd = urlopen(url, timeout=10)
    return imread(io.BytesIO(fd.read()))


def nice_scale(mapextent, maxlen=0.15):
    """Returns a nice number for a legend scale of a map.

    Parameters
    ----------
    mapextent : float
        the total extent of the map
    maxlen : float
        from 0 to 1, the maximum relative length allowed for the scale

    Examples
    --------
    >>> print(nice_scale(140))
    20.0
    >>> print(nice_scale(140, maxlen=0.5))
    50.0
    """
    d = np.array([1, 2, 5])
    e = (np.ones(12) * 10) ** (np.arange(12)-5)
    candidates = np.matmul(e[:, None],  d[None, :]).flatten()
    return np.max(candidates[candidates / mapextent <= maxlen])


def reduce(arr, factor=1, how=np.mean):
    """Reduces an array's size by a given factor.

    The reduction can be done by any reduction function (default is mean).

    Parameters
    ----------
    arr : ndarray
        an array of at least 2 dimensions (the reduction is done on the two
        last dimensions).
    factor : int
        the factor to apply for reduction (must be a divider of the original
        axis dimension!).
    how : func
        the reduction function

    Returns
    -------
    the reduced array
    """
    arr = np.asarray(arr)
    shape = list(arr.shape)
    newshape = shape[:-2] + [np.round(shape[-2] / factor).astype(int), factor,
                             np.round(shape[-1] / factor).astype(int), factor]
    return how(how(arr.reshape(*newshape), axis=len(newshape)-3),
               axis=len(newshape)-2)
